import numpy as np
import scipy.sparse as sp

from westpa.tools import Plotter

# A useful dataclass used as a wrapper for w_ipa to facilitate
# ease-of-use in ipython/jupyter notebooks/sessions.
# It basically just wraps up numpy arrays and dicts.


class WIPIDataset:
    def __init__(self, raw, key):
        self.__dict__ = {}
        self.raw = raw
        self.name = key

    def __repr__(self):
        if isinstance(self.__dict__['raw'], dict):
            return repr(self.__dir__())
        else:
            return repr(self.raw)

    def __getitem__(self, value):
        if not isinstance(value, str):
            return self.__dict__['raw'][value]
        if value in list(self.__dict__['raw'].keys()):
            return self.__dict__['raw'][value]
        elif value in list(self.__dict__.keys()):
            return self.__dict__[value]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, value):
        # Check if it's an attribute of the underlying datatype.
        # If not, just use the getitem function.
        if value in dir(self.__dict__['raw']):
            return getattr(self.__dict__['raw'], value)
        else:
            return self.__getitem__(value)

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __dir__(self):
        dict_keys = list(self.__dict__.keys())
        remove = ['raw', 'name', '__dict__', 'plotter']
        for i in remove:
            try:
                dict_keys.remove(str(i))
            except Exception:
                pass
        # We don't enforce that this is a dictionary.
        if isinstance(self.__dict__['raw'], dict):
            return sorted(set(list(self.raw.keys()) + dict_keys))
        else:
            return sorted(set(dict_keys))

    def keys(self):
        print(self.__dir__())

    # We want to override the basic math functions, now, so... this is only valid for numpy sets.
    def __add__(self, other):
        return self.__dict__['raw'] + other

    def __radd__(self, other):
        return other + self.__dict__['raw']

    def __sub__(self, other):
        return self.__dict__['raw'] - other

    def __rsub__(self, other):
        return other - self.__dict__['raw']

    def __mul__(self, other):
        return self.__dict__['raw'] * other

    def __rmul__(self, other):
        return other * self.__dict__['raw']

    def __truediv__(self, other):
        return self.__dict__['raw'] / other

    def __floordiv__(self, other):
        return self.__dict__['raw'] // other

    def __rtruediv__(self, other):
        return other / self.__dict__['raw']

    def __mod__(self, other):
        return self.__dict__['raw'] % other

    def __pow__(self, other):
        return self.__dict__['raw'] ** other

    def __lshift__(self, other):
        return self.__dict__['raw'] << other

    def __rshift__(self, other):
        return self.__dict__['raw'] >> other

    def __and__(self, other):
        return self.__dict__['raw'] & other

    def __eq__(self, other):
        return self.__dict__['raw'] == other

    def __ne__(self, other):
        return self.__dict__['raw'] != other

    def __lt__(self, other):
        return self.__dict__['raw'] < other

    def __gt__(self, other):
        return self.__dict__['raw'] > other

    def __le__(self, other):
        return self.__dict__['raw'] <= other

    def __ge__(self, other):
        return self.__dict__['raw'] >= other

    def __xor__(self, other):
        return self.__dict__['raw'] ^ other

    def __or__(self, other):
        return self.__dict__['raw'] | other

    # def __iadd__(self, other):
    #    return self.__dict__['raw'] += other
    # def __isub__(self, other):
    #    return self.__dict__['raw'] -= other
    # def __imul__(self, other):
    #    return self.__dict__['raw'] *= other
    # def __idiv__(self, other):
    #    return self.__dict__['raw'] /= other
    # def __ifloordiv__(self, other):
    #    return self.__dict__['raw'] //= other
    # def __imod__(self, other):
    #    return self.__dict__['raw'] %= other
    # def __ipow__(self, other):
    #    return self.__dict__['raw'] **= other
    # def __ilshift__(self, other):
    #    return self.__dict__['raw'] <<= other
    # def __irshift__(self, other):
    #    return self.__dict__['raw'] >>= other
    # def __iand__(self, other):
    #    return self.__dict__['raw'] &= other
    # def __ixor__(self, other):
    #    return self.__dict__['raw'] ^= other
    # def __ior__(self, other):
    #    return self.__dict__['raw'] |= other


# Similar to the above, but slightly expanded to contain information from analysis files.
class KineticsIteration:
    def __init__(self, kin_h5file, index, assign, iteration=-1):
        self.__dict__ = {}
        self.h5file = kin_h5file
        # Keys:
        _2D_h5keys = ['conditional_flux_evolution', 'rate_evolution']
        _1D_h5keys = ['state_pop_evolution', 'color_prob_evolution', 'target_flux_evolution']
        for key in _2D_h5keys:
            try:
                self.__dict__[key] = self.__2D_with_error__(key, index, assign)
            except Exception:
                self.__dict__[key] = None
        for key in _1D_h5keys:
            try:
                self.__dict__[key] = self.__1D_with_error__(key, index, assign)
            except Exception:
                self.__dict__[key] = None
        try:
            self.__dict__['total_fluxes'] = WIPIDataset(raw=np.array(self.h5file['total_fluxes']), key='total_fluxes')
            # We'll have to update this to make things better...
            # self.__dict__['total_fluxes'].plotter = Plotter(self.h5file['total_fluxes'][...], 'Total Fluxes', iteration=iteration, interface='text')
            # self.__dict__['total_fluxes'].plot = self.__dict__['total_fluxes'].plotter.plot
        except Exception:
            pass

    def __repr__(self):
        return repr(self.__dir__())

    def __getitem__(self, value):
        if value in list(self.__dict__.keys()):
            return self.__dict__[value]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, value):
        if value in list(self.__dict__.keys()):
            return self.__dict__[value]

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __dir__(self):
        dict_keys = list(self.__dict__.keys())
        # We don't want to show the plotter class; just the plot function
        remove = ['h5file', '__dict__']
        for i in remove:
            try:
                dict_keys.remove(str(i))
            except Exception:
                pass
        return sorted(set(dict_keys))

    def keys(self):
        print(self.__dir__())

    # We seriously need to rename this.
    # It's similar to the global WIPDataset, but has some nice pretty print functions.
    class __custom_dataset__:
        # This is just allow it to be indexed via properties.
        # Not a huge thing, but whatever.
        def __init__(self, raw, assign, key):
            self.__dict__ = {}
            self.raw = raw
            self.name = key
            self.assign = assign
            self.nstates = assign.attrs['nstates']
            self.dim = len(raw.shape)

        def __repr__(self):
            return repr(self.__dir__())

        def __getitem__(self, value):
            if value in self.__dict__['raw'].dtype.names:
                return self.__dict__['raw'][value]
            elif value in list(self.__dict__.keys()):
                return self.__dict__[value]

        def __setitem__(self, key, value):
            self.__dict__[key] = value

        def __getattr__(self, value):
            if value in self.__dict__['raw'].dtype.names:
                return self.__dict__['raw'][value]
            elif value in list(self.__dict__.keys()):
                return self.__dict__[value]

        def __setattr__(self, key, value):
            self.__dict__[key] = value

        def __dir__(self):
            dict_keys = list(self.__dict__.keys())
            # We don't want to show the plotter class; just the plot function
            remove = ['assign', 'dim', 'nstates', 'plotter', '__dict__']
            for i in remove:
                try:
                    dict_keys.remove(str(i))
                except Exception:
                    pass
            return sorted(set(list(self.raw.dtype.names) + dict_keys))

        def keys(self):
            print(self.__dir__())

        def _repr_pretty_(self, p, cycle):
            if self.dim == 1:
                return self._1D_repr_pretty_(p, cycle)
            if self.dim == 2:
                return self._2D_repr_pretty_(p, cycle)

        def _1D_repr_pretty_(self, p, cycle):
            # We're just using this as a way to print things in a pretty way.  They can still be indexed appropriately.
            # Stolen shamelessly from westtools/kinetics_tool.py
            maxlabellen = max(list(map(len, self.assign['state_labels'])))
            p.text('')
            p.text('{name} data:\n'.format(name=self.name))
            for istate in range(self.nstates):
                p.text(
                    '{:{maxlabellen}s}: mean={:21.15e} CI=({:21.15e}, {:21.15e}) * tau^-1\n'.format(
                        self.assign['state_labels'][istate],
                        self.raw['expected'][istate],
                        self.raw['ci_lbound'][istate],
                        self.raw['ci_ubound'][istate],
                        maxlabellen=maxlabellen,
                    )
                )
            p.text('To access data, index via the following names:\n')
            p.text(str(self.__dir__()))
            return " "

        def _2D_repr_pretty_(self, p, cycle):
            # We're just using this as a way to print things in a pretty way.  They can still be indexed appropriately.
            # Stolen shamelessly from westtools/kinetics_tool.py
            maxlabellen = max(list(map(len, self.assign['state_labels'])))
            p.text('')
            p.text('{name} data:\n'.format(name=self.name))
            for istate in range(self.nstates):
                for jstate in range(self.nstates):
                    if istate == jstate:
                        continue
                    p.text(
                        '{:{maxlabellen}s} -> {:{maxlabellen}s}: mean={:21.15e} CI=({:21.15e}, {:21.15e}) * tau^-1\n'.format(
                            self.assign['state_labels'][istate],
                            self.assign['state_labels'][jstate],
                            self.raw['expected'][istate, jstate],
                            self.raw['ci_lbound'][istate, jstate],
                            self.raw['ci_ubound'][istate, jstate],
                            maxlabellen=maxlabellen,
                        )
                    )
            p.text('To access data, index via the following names:\n')
            p.text(str(self.__dir__()))
            return " "

    def __2D_with_error__(self, h5key, index, assign):
        # Check the start and stop, calculate the block size, and index appropriately.
        # While we could try and automatically generate this above, it's a little more consistent to try it here.
        # This should show the first block for which the current iteration has contributed data.
        self.step_iter = (self.h5file[h5key]['iter_stop'][0] - self.h5file[h5key]['iter_start'][0])[1, 0]
        value = (index - self.h5file.attrs['iter_start']) // self.step_iter
        if value < 0:
            value = 0
        raw = self.h5file[h5key][value, :, :]
        error = (raw['ci_ubound'] - raw['ci_lbound']) / (2 * raw['expected'])
        # expected = raw['expected']
        raw = self.__custom_dataset__(raw, assign, h5key)
        raw.error = error
        raw.plotter = Plotter(self.h5file, h5key, iteration=value, interface='text')
        raw.plot = raw.plotter.plot
        return raw

    def __1D_with_error__(self, h5key, index, assign):
        self.step_iter = (self.h5file[h5key]['iter_stop'][0] - self.h5file[h5key]['iter_start'][0])[1]
        value = (index - self.h5file.attrs['iter_start']) // self.step_iter
        if value < 0:
            value = 0
        raw = self.h5file[h5key][value, :]
        error = (raw['ci_ubound'] - raw['ci_lbound']) / (2 * raw['expected'])
        # expected = raw['expected']
        raw = self.__custom_dataset__(raw, assign, h5key)
        raw.error = error
        raw.plotter = Plotter(self.h5file, h5key, iteration=value, interface='text')
        raw.plot = raw.plotter.plot
        return raw


class __get_data_for_iteration__:
    '''
    All interesting data from an iteration (current/past).  Whenever you change the scheme or iteration,
    this dictionary is automatically updated.  For the current iteration, it's keyed to the current seg_id.
    For the past iteration, it's keyed to the seg_id in the CURRENT iteration such that:

        w.current[X] & w.past[X]

    returns information about seg_id X in the current iteration and information on seg_ID X's PARENT in the
    preceding iteration.

    Can be indexed via a seg_id, or like a dictionary with the following keys:

        kinavg, weights, pcoord, auxdata (optional), parents, summary, seg_id, walkers, states, bins

    kinavg, states, and bins refer to the output from w_kinavg and w_assign for this iteration
    and analysis scheme.  They are NOT dynamics bins, but the bins defined in west.cfg.

    Has the following properties:

        .minweight, .maxweight

    which return all properties of the segment that matches those criteria in the selected iteration.

    If you change the analysis scheme, so, too, will the important values.
    '''

    def __init__(self, parent, value, seg_ids=None):
        '''
        Initializes and sets the correct data.
        '''
        # We've classed this so that we can override some of the normal functions and allow indexing via seg_id
        self.__dict__ = {}
        # Is this function thread safe?
        iter_group = parent.data_reader.get_iter_group(value)
        # iter_group = parent.west['iterations/iter_{num:08d}'.format(num=value)]
        self.parent = parent
        current = {}
        current['iteration'] = value
        if seg_ids is None:
            seg_ids = range(0, iter_group['seg_index']['weight'].shape[0])
        # Just make these easier to access.
        current['weights'] = iter_group['seg_index']['weight'][seg_ids]
        current['pcoord'] = iter_group['pcoord'][...][seg_ids, :, :]
        try:
            current['auxdata'] = {}
            for key in list(iter_group['auxdata'].keys()):
                current['auxdata'][key] = iter_group['auxdata'][key][...][seg_ids, :]
        except Exception:
            pass
        current['parents'] = iter_group['seg_index']['parent_id'][seg_ids]
        current['summary'] = parent.data_reader.data_manager.get_iter_summary(int(value))
        current['seg_id'] = np.array(list(range(0, iter_group['seg_index'].shape[0])))[seg_ids]
        current['walkers'] = current['summary']['n_particles']
        current['states'] = parent.assign['trajlabels'][value - 1, : current['walkers'], :][seg_ids]
        current['bins'] = parent.assign['assignments'][value - 1, : current['walkers'], :][seg_ids]
        # Calculates the bin population for this iteration.
        nbins = parent.assign['state_map'].shape[0]
        # We have to take the 'unknown' state into account
        # nstates = parent.assign['state_labels'].shape[0] + 1
        # Temporarily disabled while I sort out the fact that we shouldn't be using data from w_assign for state populations.
        # current['plot'] = Plotter(parent.direct, parent.reweight, parent.iteration, parent.assign['bin_labels'], parent.assign['state_labels'], current['populations'].states, current['populations'].bins, parent.interface)
        # Now we'll load up the results of the kinetics analysis.
        current['direct'] = KineticsIteration(parent.direct, value, parent.assign, value)
        evolution_datasets = [
            'rate_evolution',
            'conditional_flux_evolution',
            'state_pop_evolution',
            'color_prob_evolution',
            'total_fluxes',
            'target_flux_evolution',
        ]
        # We want to load these up as... oh, who knows, I suppose?
        try:
            current['reweight'] = KineticsIteration(parent.reweight, value, parent.assign, value)
            # We'll make this not a sparse matrix...
            matrix = parent.reweight['iterations/iter_{:08d}'.format(value)]
            # Assume color.
            current['instant_matrix'] = sp.coo_matrix(
                (matrix['flux'][...], (matrix['rows'][...], matrix['cols'][...])), shape=((nbins - 1) * 2, (nbins - 1) * 2)
            ).todense()
            reweighting = True
        except Exception:
            # This analysis hasn't been enabled, so we'll simply return the default error message.
            current['reweight'] = parent.reweight['rate_evolution']
            current['instant_matrix'] = parent.reweight['bin_populations']
            current['matrix'] = parent.reweight['bin_populations']
            reweighting = False
        # Check if the analysis has been enabled.  If yes, make them specify dataset dictionaries.  If not, return the thing.
        if reweighting:
            for key in evolution_datasets:
                current[key] = WIPIDataset(raw={'direct': current['direct'][key], 'reweight': current['reweight'][key]}, key='a')
        else:
            for key in evolution_datasets:
                current[key] = WIPIDataset(raw={'direct': current['direct'][key]}, key='direct')

        self.raw = current

    def __repr__(self):
        '''
        Returns the dictionary containing the iteration's values.
        '''
        return repr(self.__dir__())

    def keys(self):
        '''
        Returns the keys function of the internal dictionary.
        '''
        return list(self.__dict__['raw'].keys())

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, value):
        if value in list(self.__dict__['raw'].keys()):
            return self.__dict__['raw'][value]
        elif value in list(self.__dict__.keys()):
            return self.__dict__[value]

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __dir__(self):
        dict_keys = list(self.__dict__.keys())
        dict_keys += ['maxweight', 'minweight', 'walkers', 'aggregate_walkers', 'successful_trajectories']
        remove = ['__dict__']
        for i in remove:
            try:
                dict_keys.remove(str(i))
            except Exception:
                pass
        return sorted(set(list(self.__dict__['raw'].keys()) + dict_keys))

    @property
    def maxweight(self):
        '''
        Returns information about the segment which has the largest weight for this iteration.
        '''
        # Is there a faster or cleaner way to do this?  Ah, maybe.
        walker = np.where(self.raw['weights'] == np.max(self.raw['weights']))[0][0]
        return self.__getitem__(walker)

    @property
    def minweight(self):
        '''
        Returns information about the segment which has the smallest weight for this iteration.
        '''
        walker = np.where(self.raw['weights'] == np.min(self.raw['weights']))[0][0]
        return self.__getitem__(walker)

    @property
    def successful_trajectories(self):
        '''
        Returns which trajectories are successful.
        '''
        # walker = np.where(self.raw['weights'] == np.min(self.raw['weights']))[0][0]
        # Find where we have a transition....
        state_changes = np.where(self.raw['states'][:, :-1] != self.raw['states'][:, 1:])
        walkers = state_changes[0]
        # The index of the state change.
        new_states = state_changes[1] + 1
        old_states = state_changes[1]
        walker = {}
        for z, (i, j) in enumerate(zip(old_states, new_states)):
            # if self.raw['states'][walkers[z], i] == istate and self.raw['states'][walkers[z], j] == jstate:
            istate = self.raw['states'][walkers[z], i]
            jstate = self.raw['states'][walkers[z], j]
            # print(z,i,j, istate, jstate)
            try:
                walker[istate, jstate].append(walkers[z])
            except Exception:
                walker[istate, jstate] = [walkers[z]]

        walker = WIPIDataset(raw=walker, key=None)
        return walker

    @property
    def walkers(self):
        '''
        The number of walkers active in the current iteration.
        '''
        # Returns number of walkers for iteration X.  Assumes current iteration, but can go with different one.
        # Make this just... yeah, put this elsewhere.
        return self.parent.west['summary']['n_particles'][self.iteration - 1]

    @property
    def aggregate_walkers(self):
        return self.parent.west['summary']['n_particles'][: self.iteration].sum()

    def __getitem__(self, value):
        '''
        Responsible for handling whether this is treated like a dictionary of data sets, or an array of walker data.
        '''
        # Check to see if we're indexing via any of the active string types.  We should probably break it down via string or int, instead of 'what exists and what doesn't', but it works for now.
        # active_items = [
        # 'kinavg',
        # 'statepops',
        # 'weights',
        # 'pcoord',
        # 'auxdata',
        # 'parents',
        # 'summary',
        # 'seg_id',
        # 'walkers',
        # 'states',
        # 'bins',
        # 'populations',
        # 'plot',
        # 'instant_matrix',
        # 'kinrw',
        # 'matrix',
        # 'rwstatepops',
        # ]
        # if value in active_items:
        if isinstance(value, str):
            # This should handle everything.  Otherwise...
            try:
                return self.raw[value]
            except Exception:
                print('{} is not a valid data structure.'.format(value))
        elif isinstance(value, int) or isinstance(value, np.int64):
            # Otherwise, we assume they're trying to index for a seg_id.
            if value < self.walkers:
                current = {}
                current['plotter'] = {}
                for i in ['pcoord']:
                    current[i] = WIPIDataset(raw=self.raw[i][value, :, :], key=i)
                    current[i].plotter = Plotter(self.raw[i][value, :, :], i, iteration=self.iteration, interface='text')
                    current[i].plot = current[i].plotter.plot

                current['states'] = self.raw['states'][value, :]
                current['bins'] = self.raw['bins'][value, :]
                current['parents'] = self.raw['parents'][value]
                current['seg_id'] = self.raw['seg_id'][value]
                current['weights'] = self.raw['weights'][value]
                try:
                    current['auxdata'] = {}
                    for key in list(self.raw['auxdata'].keys()):
                        current['auxdata'][key] = self.raw['auxdata'][key][value]
                except Exception:
                    pass
                current = WIPIDataset(current, 'Segment {} in Iter {}'.format(value, self.iteration))
                return current
            else:
                print('INVALID SEG_ID {}.  SEG_ID should be less than {}.'.format(value, self.walkers))


# This handles the 'schemes', and all assorted data.
class WIPIScheme:
    def __init__(self, scheme, name, parent, settings):
        self.__dict__ = {}
        self.raw = scheme
        # self.name = parent._schemename
        self.__analysis_schemes__ = scheme
        self.iteration = parent.iteration
        self.__dict__['name'] = None
        self.__settings = settings
        # Are these necessary?  We'll try to edit these out.
        self.parent = parent
        self.data_reader = parent.data_reader

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        # Right now, this returns w.scheme, NOT necessarily what we're pulling from...
        # So you can rely on this, but it's confusing.
        if self.name is not None:
            # Set it to None, then return the original value.
            rtn_string = self.name
            self.name = None
            return rtn_string
        else:
            return str(self.scheme)

    def __getitem__(self, value):
        if not isinstance(value, str):
            for ischeme, schemename in enumerate(self.__dict__['raw'].keys()):
                if ischeme == value:
                    value = schemename
        # Check for some weird Ipython stuff.
        if '_ipython' in value:
            return self
        self.name = None
        if value in list(self.__dict__['raw'].keys()):
            # If we have it in there...
            self.name = value
            return self
        elif value in list(self.__dict__.keys()):
            self.name = value
            return self
        elif value in self.__dir__():
            self.name = value
            return self

    def __getattr__(self, value):
        return self.__getitem__(value)

    def __dir__(self):
        dict_keys = ['assign', 'direct', 'state_labels', 'bin_labels', 'west', 'reweight', 'current', 'past', 'iteration']
        if self.name is not None:
            return sorted(set(dict_keys))
        else:
            return sorted(set(self.__analysis_schemes__.keys()))

    @property
    def scheme(self):
        self.name = None
        return self.parent._schemename

    @property
    def list_schemes(self):
        '''
        Lists what schemes are configured in west.cfg file.
        Schemes should be structured as follows, in west.cfg:

        west:
          system:
            analysis:
              directory: analysis
              analysis_schemes:
                scheme.1:
                  enabled: True
                  states:
                    - label: unbound
                      coords: [[7.0]]
                    - label: bound
                      coords: [[2.7]]
                  bins:
                    - type: RectilinearBinMapper
                      boundaries: [[0.0, 2.80, 7, 10000]]
        '''
        print("The following schemes are available:")
        print("")
        for ischeme, scheme in enumerate(self.__settings['analysis_schemes']):
            print('{}. Scheme: {}'.format(ischeme, scheme))
        print("")
        print("Set via name, or via the index listed.")
        print("")
        print("Current scheme: {}".format(self.scheme))

    @property
    def iteration(self):
        return self.parent.iteration

    @property
    def assign(self):
        return self.__analysis_schemes__[str(self.name)]['assign']

    @property
    def direct(self):
        '''
        The output from w_direct.py from the current scheme.
        '''
        return self.__analysis_schemes__[str(self.name)]['direct']

    @property
    def state_labels(self):
        print("State labels and definitions!")
        for istate, state in enumerate(self.assign['state_labels']):
            print('{}: {}'.format(istate, state))
        print('{}: {}'.format(istate + 1, 'Unknown'))

    @property
    def bin_labels(self):
        print("Bin definitions! ")
        for istate, state in enumerate(self.assign['bin_labels']):
            print('{}: {}'.format(istate, state))

    @property
    def west(self):
        return self.data_reader.data_manager.we_h5file

    @property
    def reweight(self):
        # Need to fix this...
        if self.__settings['analysis_schemes'][str(self.name)]['postanalysis'] is True:
            return self.__analysis_schemes__[str(self.name)]['reweight']
        else:
            value = "This sort of analysis has not been enabled."
            current = {
                'bin_prob_evolution': value,
                'color_prob_evolution': value,
                'conditional_flux_evolution': value,
                'rate_evolution': value,
                'state_labels': value,
                'state_prob_evolution': value,
            }
            current.update({'bin_populations': value, 'iterations': value})
            return current

    @property
    def current(self):
        '''
        The current iteration.  See help for __get_data_for_iteration__
        '''
        return __get_data_for_iteration__(value=self.iteration, parent=self)

    @property
    def past(self):
        '''
        The previous iteration.  See help for __get_data_for_iteration__
        '''
        if self.iteration > 1:
            return __get_data_for_iteration__(value=self.iteration - 1, seg_ids=self.current['parents'], parent=self)
        else:
            print("The current iteration is 1; there is no past.")



import logging
import warnings

import numpy as np

from westpa.tools import WESTMasterCommand, WESTParallelTool

from westpa.tools.kinetics_tool import WESTKineticsBase, AverageCommands, generate_future
from westpa.core import h5io


from westpa.mclib import mcbs_ci_correl

from westpa.core.reweight import reweight_for_c, FluxMatrix


warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('ignore', category=RuntimeWarning)
warnings.filterwarnings('ignore', category=FutureWarning)

log = logging.getLogger('w_reweight')


def _2D_eval_block(
    iblock, start, stop, nstates, data_input, name, mcbs_alpha, mcbs_nsets, mcbs_acalpha, do_correl, mcbs_enable, estimator_kwargs
):
    # As our reweighting estimator is a weird function, we can't use the general mclib block.
    results = []
    for istate in range(nstates):
        for jstate in range(nstates):
            if istate == jstate:
                continue
            estimator_kwargs.update(dict(istate=istate, jstate=jstate, nstates=nstates))

            dataset = {'indices': np.array(list(range(start - 1, stop - 1)), dtype=np.uint16)}

            ci_res = mcbs_ci_correl(
                dataset,
                estimator=reweight_for_c,
                alpha=mcbs_alpha,
                n_sets=mcbs_nsets,
                autocorrel_alpha=mcbs_acalpha,
                subsample=(lambda x: x[0]),
                do_correl=do_correl,
                mcbs_enable=mcbs_enable,
                estimator_kwargs=estimator_kwargs,
            )
            results.append((name, iblock, istate, jstate, (start, stop) + ci_res))

    return results


def _1D_eval_block(
    iblock, start, stop, nstates, data_input, name, mcbs_alpha, mcbs_nsets, mcbs_acalpha, do_correl, mcbs_enable, estimator_kwargs
):
    # As our reweighting estimator is a weird function, we can't use the general mclib block.
    results = []
    for istate in range(nstates):
        # A little hack to make our estimator play nice, as jstate must be there.
        # For 1D datasets (state probabilities, etc), the argument isn't used in our estimator,
        # and so any variable which has the proper type is fine.
        estimator_kwargs.update(dict(istate=istate, jstate=istate, nstates=nstates))

        dataset = {'indices': np.array(list(range(start - 1, stop - 1)), dtype=np.uint16)}

        ci_res = mcbs_ci_correl(
            dataset,
            estimator=reweight_for_c,
            alpha=mcbs_alpha,
            n_sets=mcbs_nsets,
            autocorrel_alpha=mcbs_acalpha,
            subsample=(lambda x: x[0]),
            do_correl=do_correl,
            mcbs_enable=mcbs_enable,
            estimator_kwargs=estimator_kwargs,
        )
        results.append((name, iblock, istate, (start, stop) + ci_res))

    return results


def _pop_eval_block(
    iblock,
    start,
    stop,
    nstates,
    data_input,
    name,
    mcbs_alpha,
    mcbs_nsets,
    mcbs_acalpha,
    do_correl,
    mcbs_enable,
    estimator_kwargs,
    **kwargs
):
    # As our reweighting estimator is a weird function, we can't use the general mclib block.
    # A little hack to make our estimator play nice, as jstate must be there.
    # For 1D datasets (state probabilities, etc), the argument isn't used in our estimator,
    # and so any variable which has the proper type is fine.
    estimator_kwargs.update(dict(istate=0, jstate=0, nstates=nstates))

    estimator_kwargs.update({'indices': np.array(list(range(start - 1, stop - 1)), dtype=np.uint16), 'stride': 1})
    # cpdef reweight_for_c(rows, cols, obs, flux, insert, indices, nstates, nbins, state_labels, state_map, nfbins, istate, jstate, stride, bin_last_state_map, bin_state_map, return_obs, obs_threshold=1):
    # ['nbins', 'rows', 'state_map', 'nfbins', 'bin_state_map', 'cols', 'jstate', 'flux', 'istate', 'return_obs', 'bin_last_state_map', 'insert', 'nstates', 'indices', 'state_labels', 'obs']
    # cpdef reweight_for_c(stride, obs_threshold=1):

    ci_res = reweight_for_c(**estimator_kwargs)[...].reshape(-1, nstates).sum(axis=1)

    return ci_res


class RWMatrix(WESTKineticsBase, FluxMatrix):
    subcommand = 'init'
    default_kinetics_file = 'reweight.h5'
    default_output_file = 'reweight.h5'
    help_text = 'create a color-labeled transition matrix from a WESTPA simulation'
    description = '''\
Generate a colored transition matrix from a WE assignment file. The subsequent
analysis requires that the assignments are calculated using only the initial and
final time points of each trajectory segment. This may require downsampling the
h5file generated by a WE simulation. In the future w_assign may be enhanced to optionally
generate the necessary assignment file from a h5file with intermediate time points.
Additionally, this analysis is currently only valid on simulations performed under
either equilibrium or steady-state conditions without recycling target states.

-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, by default "reweight.h5") contains the
following datasets:

  ``/bin_populations`` [window, bin]
     The reweighted populations of each bin based on windows. Bins contain
     one color each, so to recover the original un-colored spatial bins,
     one must sum over all states.

  ``/iterations`` [iteration]
    *(Structured -- see below)*  Sparse matrix data from each
    iteration.  They are reconstructed and averaged within the
    w_reweight {kinetics/probs} routines so that observables may
    be calculated.  Each group contains 4 vectors of data:

      flux
        *(Floating-point)* The weight of a series of flux events
      cols
        *(Integer)* The bin from which a flux event began.
      cols
        *(Integer)* The bin into which the walker fluxed.
      obs
        *(Integer)* How many flux events were observed during this
        iteration.

-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.assignments_file = None

    def add_args(self, parser):
        cogroup = parser.add_argument_group('calculation options')
        cogroup.add_argument(
            '-s',
            '--sampling-frequency',
            dest='sampling_frequency',
            choices=['timepoint', 'iteration'],
            default='timepoint',
            help='''Observe for transition events with a lag
                             time corresponding to either each weighted ensemble
                             iteration (``iteration``) or each sub-iteration
                             timepoint (``timepoint``) (default: %(default)s).
                             For each pair of bins, store the sum of fluxes
                             in a given iteration (['flux']). Similarly,
                             for each bin, store the sum of weights in the bin
                             over all observation points for a given iteration
                             (['bin_populations']).''',
        )

    def process_args(self, args):
        self.output_file = h5io.WESTPAH5File(args.output, 'w', creating_program=True)
        self.assignments_file = h5io.WESTPAH5File(args.assignments, 'r')
        # Force a build of the transition matrix at the iteration level.
        self.sampling_frequency = 'iteration' if self.assignments_file.attrs['subsampled'] is True else args.sampling_frequency

    def go(self):
        # This function exists in FluxMatrix, and is not currently portable.
        # That is, it assumes all properties are properly initialized (self.assignments_file, etc)
        # TO DO : make it portable.
        pi = self.progress.indicator
        with pi:
            self.w_postanalysis_matrix()


class RWReweight(AverageCommands):
    help_text = 'Parent class for all reweighting routines, as they all use the same estimator code.'

    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.assignments_file = None

    def add_args(self, parser):
        cogroup = parser.add_argument_group('calculation options')
        cogroup.add_argument(
            '--obs-threshold',
            type=int,
            default=1,
            help='''The minimum number of observed transitions between two states i and j necessary to include
                             fluxes in the reweighting estimate''',
        )

    def process_args(self, args):
        self.obs_threshold = args.obs_threshold

    def accumulate_statistics(self, start_iter, stop_iter):
        '''
        This function pulls previously generated flux matrix data into memory.
        The data is assumed to exist within an HDF5 file that is available
        as a property.
        The data is kept as a single dimensional numpy array to use with the
        cython estimator.
        '''
        # This is designed to pull in the flux data...
        rows = []
        cols = []
        obs = []
        flux = []
        insert = [0] * (start_iter)

        # Actually, I'm not sure we need to start this at start_iter...
        # ... as it's keyed to the iteration, we need to make sure that the index
        # matches with the iteration (particularly for 'insert').
        # It's just easier to load all the data, although we could just start insert as a list of length
        # start_iter.
        for iiter in range(start_iter, stop_iter):
            iter_grp = self.kinetics_file['iterations']['iter_{:08d}'.format(iiter)]

            rows.append(iter_grp['rows'][...])
            cols.append(iter_grp['cols'][...])
            obs.append(iter_grp['obs'][...])
            flux.append(iter_grp['flux'][...])
            # 'insert' is the insertion point for each iteration; that is,
            # at what point do we look into the list for iteration X?
            insert.append(iter_grp['rows'][...].shape[0] + insert[-1])
        self.rows = np.concatenate(rows)
        self.cols = np.concatenate(cols)
        self.obs = np.concatenate(obs)
        self.flux = np.concatenate(flux)
        assert insert[-1] == len(self.rows)
        self.insert = np.array(insert, dtype=np.intc)

    def generate_reweight_data(self):
        '''
        This function ensures all the appropriate files are loaded, sets
        appropriate attributes necessary for all calling functions/children,
        and then calls the function to load in the flux matrix data.
        '''

        self.open_files()
        self.open_assignments()

        self.nfbins = self.kinetics_file.attrs['nrows']
        self.npts = self.kinetics_file.attrs['npts']
        self.state_map = self.assignments_file['state_map'][...]
        self.state_labels = self.assignments_file['state_labels'][...]

        # Copying this over to make it more convenient...
        self.output_file.replace_dataset('state_labels', data=self.assignments_file['state_labels'][...])
        # ... as we'll need it for ploterr.

        assert self.nstates == len(self.state_labels)
        assert self.nfbins == self.nbins * self.nstates

        start_iter, stop_iter, _ = self.iter_range.iter_start, self.iter_range.iter_stop, self.iter_range.iter_step

        # This function call loads up all the flux matrix data and stores it in memory.
        # We just give our cython estimator all the information it needs beforehand, as
        # the nature of the bootstrap means we can't know what iterations we'll actually need before
        # the routine is called.

        self.accumulate_statistics(start_iter, stop_iter)


class RWRate(RWReweight):
    subcommand = 'kinetics'
    help_text = 'Generates rate and flux values from a WESTPA simulation via reweighting.'
    default_kinetics_file = 'reweight.h5'
    default_output_file = 'reweight.h5'
    description = '''\
Calculate average rates from weighted ensemble data using the postanalysis
reweighting scheme. Bin assignments (usually "assign.h5") and pre-calculated
iteration flux matrices (usually "reweight.h5") data files must have been
previously generated using w_reweight matrix (see "w_assign --help" and
"w_reweight init --help" for information on generating these files).

-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------
The output file (-o/--output, usually "kinrw.h5") contains the following
dataset:

  /avg_rates [state,state]
    (Structured -- see below) State-to-state rates based on entire window of
    iterations selected.

  /avg_total_fluxes [state]
    (Structured -- see below) Total fluxes into each state based on entire
    window of iterations selected.

  /avg_conditional_fluxes [state,state]
    (Structured -- see below) State-to-state fluxes based on entire window of
    iterations selected.

If --evolution-mode is specified, then the following additional datasets are
available:

  /rate_evolution [window][state][state]
    (Structured -- see below). State-to-state rates based on windows of
    iterations of varying width.  If --evolution-mode=cumulative, then
    these windows all begin at the iteration specified with
    --start-iter and grow in length by --step-iter for each successive
    element. If --evolution-mode=blocked, then these windows are all of
    width --step-iter (excluding the last, which may be shorter), the first
    of which begins at iteration --start-iter.

  /target_flux_evolution [window,state]
    (Structured -- see below). Total flux into a given macro state based on
    windows of iterations of varying width, as in /rate_evolution.

  /conditional_flux_evolution [window,state,state]
    (Structured -- see below). State-to-state fluxes based on windows of
    varying width, as in /rate_evolution.

The structure of these datasets is as follows:

  iter_start
    (Integer) Iteration at which the averaging window begins (inclusive).

  iter_stop
    (Integer) Iteration at which the averaging window ends (exclusive).

  expected
    (Floating-point) Expected (mean) value of the observable as evaluated within
    this window, in units of inverse tau.

  ci_lbound
    (Floating-point) Lower bound of the confidence interval of the observable
    within this window, in units of inverse tau.

  ci_ubound
    (Floating-point) Upper bound of the confidence interval of the observable
    within this window, in units of inverse tau.

  stderr
    (Floating-point) The standard error of the mean of the observable
    within this window, in units of inverse tau.

  corr_len
    (Integer) Correlation length of the observable within this window, in units
    of tau.

Each of these datasets is also stamped with a number of attributes:

  mcbs_alpha
    (Floating-point) Alpha value of confidence intervals. (For example,
    *alpha=0.05* corresponds to a 95% confidence interval.)

  mcbs_nsets
    (Integer) Number of bootstrap data sets used in generating confidence
    intervals.

  mcbs_acalpha
    (Floating-point) Alpha value for determining correlation lengths.


-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
    '''

    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.assignments_file = None

    def w_postanalysis_reweight(self):
        '''
        This function ensures the data is ready to send in to the estimator and
        the bootstrapping routine, then does so. Much of this is simply setting
        up appropriate args and kwargs, then passing them in to the
        'run_calculation' function, which sets up future objects to send to the
        work manager. The results are returned, and then written to the
        appropriate HDF5 dataset.
        This function is specific for the rates and fluxes from the reweighting
        method.
        '''
        self.generate_reweight_data()
        pi = self.progress.indicator

        # start_iter, stop_iter, step_iter = self.iter_range.iter_start, self.iter_range.iter_stop, self.iter_range.iter_step
        # start_pts = list(range(start_iter, stop_iter, step_iter))
        start_iter = self.iter_range.iter_start
        stop_iter = self.iter_range.iter_stop

        indices = np.array(list(range(start_iter - 1, stop_iter - 1)), dtype=np.uint16)
        submit_kwargs = dict(
            pi=pi, nstates=self.nstates, start_iter=self.start_iter, stop_iter=self.stop_iter, step_iter=self.step_iter
        )

        # Due to the way our estimator is written, we're using the same dataset every time.  We're just returning different values.
        submit_kwargs['dataset'] = {'indices': indices}
        submit_kwargs['estimator_kwargs'] = {}
        submit_kwargs['estimator_kwargs'].update(
            dict(
                rows=self.rows,
                cols=self.cols,
                obs=self.obs,
                flux=self.flux,
                insert=self.insert,
                bin_last_state_map=np.tile(np.arange(self.nstates, dtype=int), self.nbins),
                bin_state_map=np.repeat(self.state_map[:-1], self.nstates),
                nfbins=self.nfbins,
                state_labels=self.state_labels,
                return_obs='R',  # Set to a default, here, but we explicitly set it later.
                state_map=self.state_map,
                nbins=self.nbins,
            )
        )

        # Calculate averages for the simulation, then report, if necessary.

        # The dataset options are what we pass on to the estimator...
        submit_kwargs['estimator_kwargs']['return_obs'] = 'R'
        avg_rates = self.run_calculation(eval_block=_2D_eval_block, name='Average Rates', dim=2, do_averages=True, **submit_kwargs)
        avg_rates['expected'] *= self.npts - 1
        avg_rates['ci_ubound'] *= self.npts - 1
        avg_rates['ci_lbound'] *= self.npts - 1
        self.output_file.replace_dataset('avg_rates', data=avg_rates[1])

        submit_kwargs['estimator_kwargs']['return_obs'] = 'F'
        avg_conditional_fluxes = self.run_calculation(
            eval_block=_2D_eval_block, name='Average Flux', dim=2, do_averages=True, **submit_kwargs
        )
        avg_conditional_fluxes['expected'] *= self.npts - 1
        avg_conditional_fluxes['ci_ubound'] *= self.npts - 1
        avg_conditional_fluxes['ci_lbound'] *= self.npts - 1
        self.output_file.replace_dataset('avg_conditional_fluxes', data=avg_conditional_fluxes[1])

        # Now, print them!

        # We've returned an average, but it still exists in a timeslice format.  So we need to return the 'last' value.
        if self.display_averages:
            self.print_averages(avg_conditional_fluxes[1], '\nfluxes from state to state:', dim=2)
            self.print_averages(avg_rates[1], '\nrates from state to state:', dim=2)

        # Do a bootstrap evolution.

        submit_kwargs['estimator_kwargs']['return_obs'] = 'R'
        rate_evol = self.run_calculation(eval_block=_2D_eval_block, name='Rate Evolution', dim=2, **submit_kwargs)
        rate_evol['expected'] *= self.npts - 1
        rate_evol['ci_ubound'] *= self.npts - 1
        rate_evol['ci_lbound'] *= self.npts - 1
        self.output_file.replace_dataset('rate_evolution', data=rate_evol, shuffle=True, compression=9)

        submit_kwargs['estimator_kwargs']['return_obs'] = 'F'
        flux_evol = self.run_calculation(eval_block=_2D_eval_block, name='Conditional Flux Evolution', dim=2, **submit_kwargs)
        flux_evol['expected'] *= self.npts - 1
        flux_evol['ci_ubound'] *= self.npts - 1
        flux_evol['ci_lbound'] *= self.npts - 1
        self.output_file.replace_dataset('conditional_flux_evolution', data=rate_evol, shuffle=True, compression=9)

    def go(self):
        pi = self.progress.indicator
        with pi:
            self.w_postanalysis_reweight()


class RWStateProbs(RWReweight):
    subcommand = 'probs'
    help_text = 'Calculates color and state probabilities via reweighting.'
    default_kinetics_file = 'reweight.h5'
    description = '''\
Calculate average populations from weighted ensemble data using the postanalysis
reweighting scheme. Bin assignments (usually "assign.h5") and pre-calculated
iteration flux matrices (usually "reweight.h5") data files must have been
previously generated using w_reweight matrix (see "w_assign --help" and
"w_reweight init --help" for information on generating these files).

-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, usually "direct.h5") contains the following
dataset:

  /avg_state_probs [state]
    (Structured -- see below) Population of each state across entire
    range specified.

  /avg_color_probs [state]
    (Structured -- see below) Population of each ensemble across entire
    range specified.

If --evolution-mode is specified, then the following additional datasets are
available:

  /state_pop_evolution [window][state]
    (Structured -- see below). State populations based on windows of
    iterations of varying width.  If --evolution-mode=cumulative, then
    these windows all begin at the iteration specified with
    --start-iter and grow in length by --step-iter for each successive
    element. If --evolution-mode=blocked, then these windows are all of
    width --step-iter (excluding the last, which may be shorter), the first
    of which begins at iteration --start-iter.

  /color_prob_evolution [window][state]
    (Structured -- see below). Ensemble populations based on windows of
    iterations of varying width.  If --evolution-mode=cumulative, then
    these windows all begin at the iteration specified with
    --start-iter and grow in length by --step-iter for each successive
    element. If --evolution-mode=blocked, then these windows are all of
    width --step-iter (excluding the last, which may be shorter), the first
    of which begins at iteration --start-iter.

The structure of these datasets is as follows:

  iter_start
    (Integer) Iteration at which the averaging window begins (inclusive).

  iter_stop
    (Integer) Iteration at which the averaging window ends (exclusive).

  expected
    (Floating-point) Expected (mean) value of the observable as evaluated within
    this window, in units of inverse tau.

  ci_lbound
    (Floating-point) Lower bound of the confidence interval of the observable
    within this window, in units of inverse tau.

  ci_ubound
    (Floating-point) Upper bound of the confidence interval of the observable
    within this window, in units of inverse tau.

  stderr
    (Floating-point) The standard error of the mean of the observable
    within this window, in units of inverse tau.

  corr_len
    (Integer) Correlation length of the observable within this window, in units
    of tau.


Each of these datasets is also stamped with a number of attributes:

  mcbs_alpha
    (Floating-point) Alpha value of confidence intervals. (For example,
    *alpha=0.05* corresponds to a 95% confidence interval.)

  mcbs_nsets
    (Integer) Number of bootstrap data sets used in generating confidence
    intervals.

  mcbs_acalpha
    (Floating-point) Alpha value for determining correlation lengths.


-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def w_postanalysis_stateprobs(self):
        '''
        This function ensures the data is ready to send in to the estimator and
        the bootstrapping routine, then does so. Much of this is simply setting
        up appropriate args and kwargs, then passing them in to the
        'run_calculation' function, which sets up future objects to send to the
        work manager. The results are returned, and then written to the
        appropriate HDF5 dataset.
        This function is specific for the color (steady-state) and macrostate
        probabilities from the reweighting method.
        '''
        self.generate_reweight_data()
        pi = self.progress.indicator

        start_iter, stop_iter, step_iter = self.iter_range.iter_start, self.iter_range.iter_stop, self.iter_range.iter_step
        start_pts = list(range(start_iter, stop_iter, step_iter))
        indices = np.array(list(range(start_iter - 1, stop_iter - 1, step_iter)), dtype=np.uint16)
        submit_kwargs = dict(
            pi=pi, nstates=self.nstates, start_iter=self.start_iter, stop_iter=self.stop_iter, step_iter=self.step_iter
        )

        # Due to the way our estimator is written, we're using the same dataset every time.  We're just returning different values.
        submit_kwargs['dataset'] = {'indices': indices}
        submit_kwargs['estimator_kwargs'] = {}
        submit_kwargs['estimator_kwargs'].update(
            dict(
                rows=self.rows,
                cols=self.cols,
                obs=self.obs,
                flux=self.flux,
                insert=self.insert,
                bin_last_state_map=np.tile(np.arange(self.nstates, dtype=int), self.nbins),
                bin_state_map=np.repeat(self.state_map[:-1], self.nstates),
                nfbins=self.nfbins,
                state_labels=self.state_labels,
                return_obs='C',  # Set to a default, but we explicitly set it later.
                state_map=self.state_map,
                nbins=self.nbins,
            )
        )

        # Calculate averages for the simulation, then report, if necessary.

        # The dataset options are what we pass on to the estimator...
        submit_kwargs['estimator_kwargs']['return_obs'] = 'C'
        avg_color_probs = self.run_calculation(
            eval_block=_1D_eval_block, name='Average Color (Ensemble) Probability', dim=1, do_averages=True, **submit_kwargs
        )
        self.output_file.replace_dataset('avg_color_probs', data=avg_color_probs[1])

        submit_kwargs['estimator_kwargs']['return_obs'] = 'S'
        avg_state_probs = self.run_calculation(
            eval_block=_1D_eval_block, name='Average State Probability', dim=1, do_averages=True, **submit_kwargs
        )
        self.output_file.replace_dataset('avg_state_probs', data=avg_state_probs[1])

        # Now, print them!

        # We've returned an average, but it still exists in a timeslice format.  So we need to return the 'last' value.
        if self.display_averages:
            self.print_averages(avg_color_probs[1], '\naverage color probabilities:', dim=1)
            self.print_averages(avg_state_probs[1], '\naverage state probabilities:', dim=1)

        # Do a bootstrap evolution.

        submit_kwargs['estimator_kwargs']['return_obs'] = 'C'
        color_evol = self.run_calculation(
            eval_block=_1D_eval_block, name='Color (Ensemble) Probability Evolution', dim=1, **submit_kwargs
        )
        self.output_file.replace_dataset('color_prob_evolution', data=color_evol, shuffle=True, compression=9)

        submit_kwargs['estimator_kwargs']['return_obs'] = 'S'
        state_evol = self.run_calculation(eval_block=_1D_eval_block, name='State Probability Evolution', dim=1, **submit_kwargs)
        self.output_file.replace_dataset('state_pop_evolution', data=state_evol, shuffle=True, compression=9)

        # Little different, here.  Do a non-bootstrap evolution.
        # Just use the work manager and the estimator without the bootstrap code, which is problematic for a dataset like this.
        # We'll be adding this in later.
        if False:
            submit_kwargs['estimator_kwargs']['return_obs'] = 'P'
            submit_kwargs['pi'] = None
            futures = []
            bin_pops = []
            for iblock, start in enumerate(start_pts):
                stop = min(start + step_iter, stop_iter)
                # TODO: below line references uninitialized variable `do_averages`. Removed for now to satisfy
                # flake8 since code never runs because of `if False`. Needs to be dealt with if this changes
                # if self.evolution_mode == 'cumulative' or do_averages is True:
                if self.evolution_mode == 'cumulative':
                    windowsize = int(self.evol_window_frac * (stop - start_iter))
                    block_start = max(start_iter, stop - windowsize)
                else:  # self.evolution_mode == 'blocked'
                    block_start = start
                future_kwargs = dict(
                    iblock=iblock,
                    start=block_start,
                    stop=stop,
                    # nstates=self.nstates,
                    mcbs_alpha=self.mcbs_alpha,
                    mcbs_nsets=self.mcbs_nsets,
                    mcbs_acalpha=self.mcbs_acalpha,
                    do_correl=self.do_correl,
                    name='Bin Population Evolution',
                    mcbs_enable=self.mcbs_enable,
                    data_input={},
                    **submit_kwargs
                )
                # print(future_kwargs)
                futures.append(generate_future(self.work_manager, 'Bin Pop Evolution', _pop_eval_block, future_kwargs))
            # bin_evol = self.run_calculation(eval_block=_pop_eval_block, name='Bin Population Evolution', dim=1, **submit_kwargs)
            # bin_evol = bin_evol.reshape(-1, 2)
            for future in self.work_manager.as_completed(futures):
                # pi.progress += iblock / step_iter
                bin_pops.append(future.get_result(discard=True))
            hist = self.output_file.replace_dataset('histograms', data=bin_pops, shuffle=True, compression=9)
            hist.attrs['iter_start'] = start_iter
            hist.attrs['iter_stop'] = stop_iter
            import ast

            binbounds = []
            midpoints = []
            for i in self.assignments_file['bin_labels']:
                binbounds.append(ast.literal_eval(i)[0][0])
            # This is probably crap, so...
            binbounds.append(ast.literal_eval(self.assignments_file['bin_labels'][-1])[0][1])
            # binbounds.append(binbounds[-1]+1)
            for i in range(1, len(binbounds)):
                midpoints.append(((binbounds[i] - binbounds[i - 1]) / 2) + binbounds[i - 1])
            self.output_file.replace_dataset('binbounds_0', data=binbounds, shuffle=True, compression=9)
            self.output_file.replace_dataset('midpoints_0', data=midpoints, shuffle=True, compression=9)
            self.output_file.replace_dataset('n_iter', data=list(range(start_iter, stop_iter)), shuffle=True, compression=9)

    def go(self):
        pi = self.progress.indicator
        with pi:
            self.w_postanalysis_stateprobs()


class RWAll(RWMatrix, RWStateProbs, RWRate):
    subcommand = 'all'
    help_text = 'Runs the full suite, including the generation of the flux matrices.'
    default_kinetics_file = 'reweight.h5'
    default_output_file = 'reweight.h5'
    description = '''\
A convenience function to run init/kinetics/probs. Bin assignments,
including macrostate definitions, are required. (See
"w_assign --help" for more information).

For more information on the individual subcommands this subs in for, run
w_reweight {init/kinetics/probs} --help.

-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def go(self):
        pi = self.progress.indicator
        with pi:
            self.w_postanalysis_matrix()
            self.w_postanalysis_reweight()
            self.w_postanalysis_stateprobs()


# Just a convenience class to average the observables.
class RWAverage(RWStateProbs, RWRate):
    subcommand = 'average'
    help_text = 'Averages and returns fluxes, rates, and color/state populations.'
    default_kinetics_file = 'reweight.h5'
    default_output_file = 'reweight.h5'
    description = '''\
A convenience function to run kinetics/probs. Bin assignments,
including macrostate definitions, are required. (See
"w_assign --help" for more information).

For more information on the individual subcommands this subs in for, run
w_reweight {kinetics/probs} --help.

-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def go(self):
        pi = self.progress.indicator
        with pi:
            self.w_postanalysis_reweight()
            self.w_postanalysis_stateprobs()


class WReweight(WESTMasterCommand, WESTParallelTool):
    prog = 'w_reweight'
    subcommands = [RWMatrix, RWAverage, RWRate, RWStateProbs, RWAll]
    subparsers_title = 'reweighting kinetics analysis scheme'


def entry_point():
    warnings.warn('{} is being deprecated.  Please use WESS, WEED or the haMSM plugin instead.'.format(WReweight.prog))
    WReweight().main()


if __name__ == '__main__':
    entry_point()



import westpa

import logging

log = logging.getLogger(__name__)


def get_pcoord(state):
    log.debug('getting progress coordinate for {!r}'.format(state))
    propagator = westpa.rc.get_propagator()
    propagator.get_pcoord(state)
    return state


def gen_istate(basis_state, initial_state):
    log.debug('generating initial state from {!r} (into {!r})'.format(basis_state, initial_state))
    propagator = westpa.rc.get_propagator()
    propagator.update_basis_initial_states([basis_state], [initial_state])
    propagator.gen_istate(basis_state, initial_state)
    return basis_state, initial_state


def prep_iter(n_iter, segments):
    log.debug('propagator.prepare_iteration(...)')
    propagator = westpa.rc.get_propagator()
    propagator.clear_basis_initial_states()
    propagator.prepare_iteration(n_iter, segments)


def post_iter(n_iter, segments):
    log.debug('propagator.finalize_iteration(...)')
    propagator = westpa.rc.get_propagator()
    propagator.finalize_iteration(n_iter, segments)


def propagate(basis_states, initial_states, segments):
    propagator = westpa.rc.get_propagator()
    propagator.update_basis_initial_states(basis_states, initial_states)
    outgoing_ids = [segment.seg_id for segment in segments]
    incoming_segments = {segment.seg_id: segment for segment in propagator.propagate(segments)}
    if log.isEnabledFor(logging.DEBUG):
        log.debug('propagated {:d} segments'.format(len(incoming_segments)))
    return [incoming_segments[seg_id] for seg_id in outgoing_ids]



import logging

from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
log = logging.getLogger("msm_we")
log.setLevel(logging.DEBUG)

import numpy as np


def processCoordinates(self, coords):
    log.debug("Processing coordinates")

    if self.dimReduceMethod == "none":
        nC = np.shape(coords)
        nC = nC[0]
        data = coords.reshape(nC, 3 * self.nAtoms)
        return data

    if self.dimReduceMethod == "pca" or self.dimReduceMethod == "vamp":
        ### NaCl dimensionality reduction
        log.warning("Hardcoded selection: Doing dim reduction for Na, Cl. This is only for testing!")
        indNA = self.reference_structure.topology.select("element Na")
        indCL = self.reference_structure.topology.select("element Cl")

        diff = np.subtract(coords[:, indNA], coords[:, indCL])

        dist = np.array(np.sqrt(np.mean(np.power(diff, 2), axis=-1)))

        return dist


log.info("Loading user-override functions for modelWE")



'''tools -- classes for implementing command-line tools for WESTPA'''

from .core import WESTTool, WESTParallelTool, WESTToolComponent, WESTSubcommand, WESTMasterCommand, WESTMultiTool
from .data_reader import WESTDataReader, WESTDSSynthesizer, WESTWDSSynthesizer
from .iter_range import IterRangeSelection
from .selected_segs import SegSelector
from .binning import BinMappingComponent, mapper_from_dict
from .progress import ProgressIndicatorComponent
from .plot import Plotter
from .wipi import WIPIDataset, KineticsIteration, __get_data_for_iteration__, WIPIScheme


__all__ = [
    'WESTTool',
    'WESTParallelTool',
    'WESTToolComponent',
    'WESTSubcommand',
    'WESTMasterCommand',
    'WESTMultiTool',
    'WESTDataReader',
    'WESTDSSynthesizer',
    'WESTWDSSynthesizer',
    'IterRangeSelection',
    'SegSelector',
    'BinMappingComponent',
    'mapper_from_dict',
    'ProgressIndicatorComponent',
    'Plotter',
    'WIPIDataset',
    'KineticsIteration',
    '__get_data_for_iteration__',
    'WIPIScheme',
]



import numpy as np

from westpa.core.segment import Segment


class BasisState:
    '''Describes an basis (micro)state. These basis states are used to generate
    initial states for new trajectories, either at the beginning of the simulation
    (i.e. at w_init) or due to recycling.

    :ivar state_id:     Integer identifier of this state, usually set by the
                        data manager.
    :ivar label:        A descriptive label for this microstate (may be empty)
    :ivar probability:  Probability of this state to be selected when creating a
                        new trajectory.
    :ivar pcoord:       The representative progress coordinate of this state.

    :ivar auxref:       A user-provided (string) reference for locating data associated
                        with this state (usually a filesystem path).
    '''

    def __init__(self, label, probability, pcoord=None, auxref=None, state_id=None):
        self.label = str(label, encoding="UTF-8") if type(label) is bytes else label
        self.probability = probability
        self.pcoord = np.atleast_1d(pcoord)
        self.auxref = auxref
        self.state_id = state_id
        self.data = {}

    def __repr__(self):
        return '{} state_id={self.state_id!r} label={self.label!r} prob={self.probability!r} pcoord={self.pcoord!r}>'.format(
            object.__repr__(self)[:-1], self=self
        )

    @classmethod
    def states_to_file(cls, states, fileobj):
        '''Write a file defining basis states, which may then be read by `states_from_file()`.'''

        if isinstance(fileobj, str):
            fileobj = open(fileobj, 'wt')

        max_label_len = max(8, max(len(state.label or '') for state in states))
        max_auxref_len = max(8, max(len(state.auxref or '') for state in states))
        fmt = (
            '{state.label:<{max_label_len}s}    {state.probability:12.7g}    {auxref_str:<{max_auxref_len}s}'
            '    # state_id={state_id_str:s}    pcoord={pcoord_str}\n'
        )
        fileobj.write(
            '# {:{max_label_len}s}    {:>12s}    {:{max_auxref_len}s}\n'.format(
                'Label', 'Probability', 'Auxref', max_label_len=max_label_len - 2, max_auxref_len=max_auxref_len
            )
        )
        for state in states:
            state_id_str = str(state.state_id) if state.state_id is not None else 'None'
            pcoord_str = str(list(state.pcoord))
            auxref_str = 'None' if state.auxref == b'' or state.auxref == "b''" else str(state.auxref)
            fileobj.write(
                fmt.format(
                    state=state,
                    pcoord_str=pcoord_str,
                    auxref_str=auxref_str,
                    state_id_str=state_id_str,
                    max_label_len=max_label_len,
                    max_auxref_len=max_auxref_len,
                )
            )

    @classmethod
    def states_from_file(cls, statefile):
        '''Read a file defining basis states.  Each line defines a state, and contains a label, the probability,
        and optionally a data reference, separated by whitespace, as in::

            unbound    1.0

        or::

            unbound_0    0.6        state0.pdb
            unbound_1    0.4        state1.pdb

        '''
        states = []
        lineno = 0

        try:
            open_statefile = open(statefile, 'rt')
        except TypeError:
            open_statefile = statefile

        for line in open_statefile:
            lineno += 1

            # remove comment portion
            line = line.partition('#')[0].strip()
            if not line:
                continue

            fields = line.split()
            label = fields[0]
            try:
                probability = float(fields[1])
            except ValueError:
                raise ValueError('invalid probability ({!r}) {} line {:d}'.format(fields[1], statefile, lineno))

            try:
                auxref = fields[2].strip()
            except IndexError:
                auxref = None

            states.append(cls(state_id=None, probability=probability, label=label, auxref=auxref))

        try:
            open_statefile.close()
        except Exception:
            pass

        return states

    def as_numpy_record(self):
        '''Return the data for this state as a numpy record array.'''

        from westpa.core.data_manager import vstr_dtype, weight_dtype, seg_id_dtype

        bstate_dtype = np.dtype(
            [
                ('state_id', seg_id_dtype),
                ('probability', weight_dtype),
                ('pcoord', self.pcoord.dtype, (len(self.pcoord),)),
                ('label', vstr_dtype),
                ('auxref', vstr_dtype),
            ]
        )
        bstaterec = np.array(
            [(self.state_id, self.probability, self.pcoord, self.label or '', self.auxref or '')], dtype=bstate_dtype
        )[0]
        return bstaterec


class InitialState:
    '''Describes an initial state for a new trajectory. These are generally constructed by
    appropriate modification of a basis state.

    :ivar state_id:         Integer identifier of this state, usually set by the
                            data manager.
    :ivar basis_state_id:   Identifier of the basis state from which this state was
                            generated, or None.
    :ivar basis_state:      The `BasisState` from which this state was generated, or None.
    :ivar iter_created:     Iteration in which this state was generated (0 for
                            simulation initialization).
    :ivar iter_used:        Iteration in which this state was used to initiate a
                            trajectory (None for unused).
    :ivar istate_type:      Integer describing the type of this initial state
                            (ISTATE_TYPE_BASIS for direct use of a basis state,
                            ISTATE_TYPE_GENERATED for a state generated from a basis state,
                            ISTATE_TYPE_RESTART for a state corresponding to the endpoint
                            of a segment in another simulation, or
                            ISTATE_TYPE_START for a state generated from a start state).
    :ivar istate_status:    Integer describing whether this initial state has been properly
                            prepared.
    :ivar pcoord:           The representative progress coordinate of this state.
    '''

    ISTATE_TYPE_UNSET = 0
    ISTATE_TYPE_BASIS = 1
    ISTATE_TYPE_GENERATED = 2
    ISTATE_TYPE_RESTART = 3
    ISTATE_TYPE_START = 4

    ISTATE_UNUSED = 0

    ISTATE_STATUS_PENDING = 0
    ISTATE_STATUS_PREPARED = 1
    ISTATE_STATUS_FAILED = 2

    istate_types = {}
    istate_type_names = {}

    istate_statuses = {}
    istate_status_names = {}

    def __init__(
        self,
        state_id,
        basis_state_id,
        iter_created,
        iter_used=None,
        istate_type=None,
        istate_status=None,
        pcoord=None,
        basis_state=None,
        basis_auxref=None,
    ):
        self.state_id = state_id
        self.basis_state_id = basis_state_id
        self.basis_state = basis_state
        self.istate_type = istate_type
        self.istate_status = istate_status
        self.iter_created = iter_created
        self.iter_used = iter_used
        self.pcoord = np.atleast_1d(pcoord)
        self.basis_auxref = basis_auxref
        self.data = {}

    def __repr__(self):
        return '{} state_id={self.state_id!r} istate_type={self.istate_type!r} basis_state_id={self.basis_state_id!r} iter_created={self.iter_created!r} pcoord={self.pcoord!r}>'.format(
            object.__repr__(self)[:-1], self=self
        )

    def as_numpy_record(self):
        from westpa.core.data_manager import seg_id_dtype, istate_type_dtype, istate_status_dtype, vstr_dtype

        istate_dtype = np.dtype(
            [
                ('state_id', seg_id_dtype),
                ('basis_state_id', seg_id_dtype),
                ('iter_created', np.uint),
                ('iter_used', np.uint),
                ('istate_type', istate_type_dtype),
                ('istate_status', istate_status_dtype),
                ('basis_auxref', vstr_dtype),
                ('pcoord', self.pcoord.dtype, (len(self.pcoord),)),
            ]
        )
        return np.array(
            [
                (
                    self.state_id,
                    self.basis_state_id or 0,
                    self.iter_created or 0,
                    self.iter_used or 0,
                    self.istate_type or 0,
                    self.istate_status or 0,
                    self.basis_auxref or '',
                    self.pcoord,
                )
            ],
            dtype=istate_dtype,
        )[0]


InitialState.istate_statuses.update(
    {_attr: getattr(InitialState, _attr) for _attr in dir(InitialState) if _attr.startswith('ISTATE_STATUS_')}
)
InitialState.istate_types.update(
    {_attr: getattr(InitialState, _attr) for _attr in dir(InitialState) if _attr.startswith('ISTATE_TYPE_')}
)
InitialState.istate_status_names.update({istate_status: name for name, istate_status in InitialState.istate_statuses.items()})
InitialState.istate_type_names.update({istate_type: name for name, istate_type in InitialState.istate_types.items()})


class TargetState:
    '''Describes a target state.

    :ivar state_id:     Integer identifier of this state, usually set by the
                        data manager.
    :ivar label:        A descriptive label for this microstate (may be empty)
    :ivar pcoord: The representative progress coordinate of this state.

    '''

    def __init__(self, label, pcoord, state_id=None):
        self.label = label
        self.pcoord = np.atleast_1d(pcoord)
        self.state_id = state_id

    def __repr__(self):
        return '{} state_id={self.state_id!r} label={self.label!r} pcoord={self.pcoord!r}>'.format(
            object.__repr__(self)[:-1], self=self
        )

    @classmethod
    def states_to_file(cls, states, fileobj):
        '''Write a file defining basis states, which may then be read by `states_from_file()`.'''

        if isinstance(fileobj, str):
            fileobj = open(fileobj, 'wt')

        max_label_len = max(8, max(len(state.label or '') for state in states))

        fileobj.write('# {:{max_label_len}s}    {:s}\n'.format('Label', 'Pcoord', max_label_len=max_label_len - 2))
        for state in states:
            pcoord_str = '    '.join(str(field) for field in state.pcoord)
            fileobj.write('{:{max_label_len}s}    {:s}\n'.format(state.label, pcoord_str, max_label_len=max_label_len))

    @classmethod
    def states_from_file(cls, statefile, dtype):
        '''Read a file defining target states.  Each line defines a state, and contains a label followed
        by a representative progress coordinate value, separated by whitespace, as in::

            bound     0.02

        for a single target and one-dimensional progress coordinates or::

            bound    2.7    0.0
            drift    100    50.0

        for two targets and a two-dimensional progress coordinate.
        '''

        labels = []
        pcoord_values = []

        try:
            open_statefile = open(statefile, 'r')
        except TypeError:
            open_statefile = statefile

        for line in open_statefile:
            line = line.partition('#')[0].strip()
            if not line:
                continue

            fields = line.split()
            labels.append(fields[0])
            pcoord_values.append(np.array(list(map(dtype, fields[1:])), dtype=dtype))

        try:
            open_statefile.close()
        except Exception:
            pass

        return [cls(label=label, pcoord=pcoord) for label, pcoord in zip(labels, pcoord_values)]


def pare_basis_initial_states(basis_states, initial_states, segments=None):
    '''Given iterables of basis and initial states (and optionally segments that use them),
    return minimal sets (as in __builtins__.set) of states needed to describe the history of the given
    segments an initial states.'''

    bstatemap = {state.state_id: state for state in basis_states}
    istatemap = {state.state_id: state for state in initial_states}

    if segments is not None:
        segments = list(segments)
        return_istates = set(
            istatemap[segment.initial_state_id] for segment in segments if segment.initpoint_type == Segment.SEG_INITPOINT_NEWTRAJ
        )
    else:
        return_istates = set(initial_states)

    return_bstates = set(
        bstatemap[istate.basis_state_id]
        for istate in return_istates
        if istate.istate_type != InitialState.ISTATE_TYPE_RESTART and istate.istate_type != InitialState.ISTATE_TYPE_START
    )

    return return_bstates, return_istates


def return_state_type(state_obj):
    '''Convinience function for returning the state ID and type of the state_obj pointer'''

    if isinstance(state_obj, Segment):
        return type(state_obj).__name__, state_obj.seg_id
    elif isinstance(state_obj, InitialState):
        return type(state_obj).__name__, state_obj.state_id
    elif isinstance(state_obj, BasisState):
        return type(state_obj).__name__, state_obj.state_id
    else:
        return 'Unknown', float('inf')



import logging
import types

import numpy as np

import westpa
from westpa.core import extloader
from westpa.core.binning import VoronoiBinMapper
from westpa.core.yamlcfg import check_bool, ConfigItemMissing

from westpa.westext.stringmethod import WESTStringMethod, DefaultStringMethod


log = logging.getLogger(__name__)


class StringDriver:
    def __init__(self, sim_manager, plugin_config):
        super().__init__()

        if not sim_manager.work_manager.is_master:
            return

        self.sim_manager = sim_manager
        self.data_manager = sim_manager.data_manager
        self.system = sim_manager.system

        # Parameters from config file
        self.windowsize = plugin_config.get('windowsize', 10)
        self.update_interval = plugin_config.get('update_interval', 10)
        self.initial_update = plugin_config.get('initial_update', 20)
        self.priority = plugin_config.get('priority', 0)

        self.write_avg_pos = check_bool(plugin_config.get('write_avgpos', True))
        self.do_update = check_bool(plugin_config.get('do_update', True))
        self.init_from_data = check_bool(plugin_config.get('init_from_data', True))

        self.dfunc = self.get_dfunc_method(plugin_config)

        # Load method to calculate average position in a bin
        # If the method is defined in an external module, correctly bind it
        ap = self.get_avgpos_method(plugin_config)
        if hasattr(ap, 'im_class'):
            self.get_avgpos = ap
        else:
            self.get_avgpos = types.MethodType(ap, self)

        # Get initial set of string centers
        centers = self.get_initial_centers()

        try:
            sm_params = self.system.sm_params
        except AttributeError as e:
            log.error(
                'String Driver Error: system does not define sm_params. \
                        This is required and should be added to the system definition; {}'.format(
                    e
                )
            )
            raise

        # Initialize the string
        str_method = self.get_string_method(plugin_config)

        try:
            self.strings = str_method(centers, **sm_params)
        except (TypeError, AssertionError) as e:
            log.error('String Driver Error: Failed during initialization of string method: {}'.format(e))
            raise

        # Update the BinMapper
        self.update_bin_mapper()

        # Register callback
        sim_manager.register_callback(sim_manager.prepare_new_iteration, self.prepare_new_iteration, self.priority)

        westpa.rc.pstatus('-westext.stringmethod -----------------\n')
        westpa.rc.pstatus('windowsize: {}\n'.format(self.windowsize))
        westpa.rc.pstatus('update interval: {}\n'.format(self.update_interval))
        westpa.rc.pstatus('initial update: {}\n'.format(self.initial_update))
        westpa.rc.pstatus('priority: {}\n'.format(self.priority))
        westpa.rc.pstatus('write average positions: {}\n'.format(self.write_avg_pos))
        westpa.rc.pstatus('do update: {}\n'.format(self.do_update))
        westpa.rc.pstatus('initialize from WE data: {}\n'.format(self.init_from_data))
        westpa.rc.pstatus('----------------------------------------\n')
        westpa.rc.pflush()

    def dfunc(self):
        raise NotImplementedError

    def get_avgpos(self, n_iter):
        raise NotImplementedError

    def get_dfunc_method(self, plugin_config):
        try:
            methodname = plugin_config['dfunc_method']
        except KeyError:
            raise ConfigItemMissing('dfunc_method')

        dfunc_method = extloader.get_object(methodname)

        log.info('loaded stringmethod dfunc method {!r}'.format(dfunc_method))

        return dfunc_method

    def get_avgpos_method(self, plugin_config):
        try:
            methodname = plugin_config['avgpos_method']
        except KeyError:
            raise ConfigItemMissing('avgpos_method')

        if methodname.lower() == 'cartesian':
            avgpos_method = self.avgpos_cartesian
        else:
            avgpos_method = extloader.get_object(methodname)

        log.info('loaded stringmethod avgpos method {!r}'.format(avgpos_method))

        return avgpos_method

    def get_string_method(self, plugin_config):
        try:
            methodname = plugin_config['string_method']
        except KeyError:
            raise ConfigItemMissing('string_method')

        if methodname.lower() == 'default':
            str_method = DefaultStringMethod
        else:
            str_method = extloader.get_object(methodname)

        assert issubclass(str_method, WESTStringMethod)
        log.debug('loaded stringmethod string method {!r}'.format(str_method))

        return str_method

    def get_initial_centers(self):
        self.data_manager.open_backing()

        with self.data_manager.lock:
            n_iter = max(self.data_manager.current_iteration - 1, 1)
            iter_group = self.data_manager.get_iter_group(n_iter)

            # First attempt to initialize string from data rather than system
            centers = None
            if self.init_from_data:
                log.info('Attempting to initialize stringmethod from data')

                try:
                    binhash = iter_group.attrs['binhash'].encode()
                    bin_mapper = self.data_manager.get_bin_mapper(binhash)

                    centers = bin_mapper.centers

                except Exception:
                    log.warning('Initializing string centers from data failed; Using definition in system instead.')
                    centers = self.system.bin_mapper.centers
            else:
                log.info('Initializing string centers from system definition')
                centers = self.system.bin_mapper.centers

        self.data_manager.close_backing()

        return centers

    def update_bin_mapper(self):
        '''Update the bin_mapper using the current string'''

        westpa.rc.pstatus('westext.stringmethod: Updating bin mapper\n')
        westpa.rc.pflush()

        try:
            dfargs = getattr(self.system, 'dfargs', None)
            dfkwargs = getattr(self.system, 'dfkwargs', None)
            self.system.bin_mapper = VoronoiBinMapper(self.dfunc, self.strings.centers, dfargs=dfargs, dfkwargs=dfkwargs)
        except (ValueError, TypeError) as e:
            log.error('StringDriver Error: Failed updating the bin mapper: {}'.format(e))
            raise

    def avgpos_cartesian(self, n_iter):
        '''Get average position of replicas in each bin as of n_iter for the
        the user selected update interval'''

        nbins = self.system.bin_mapper.nbins
        ndim = self.system.pcoord_ndim

        avg_pos = np.zeros((nbins, ndim), dtype=self.system.pcoord_dtype)
        sum_bin_weight = np.zeros((nbins,), dtype=self.system.pcoord_dtype)

        start_iter = max(n_iter - min(self.windowsize, n_iter), 2)
        stop_iter = n_iter + 1

        for n in range(start_iter, stop_iter):
            with self.data_manager.lock:
                iter_group = self.data_manager.get_iter_group(n)
                seg_index = iter_group['seg_index'][...]

                pcoords = iter_group['pcoord'][:, -1, :]  # Only read final point
                bin_indices = self.system.bin_mapper.assign(pcoords)
                weights = seg_index['weight']

                pcoord_w = pcoords * weights[:, np.newaxis]
                uniq_indices = np.unique(bin_indices)

                for indx in uniq_indices:
                    avg_pos[indx, :] += pcoord_w[bin_indices == indx].sum(axis=0)

                sum_bin_weight += np.bincount(bin_indices.astype(int), weights=weights, minlength=nbins)

        # Some bins might have zero samples so exclude to avoid divide by zero
        occ_ind = np.nonzero(sum_bin_weight)
        avg_pos[occ_ind] /= sum_bin_weight[occ_ind][:, np.newaxis]

        return avg_pos, sum_bin_weight

    def prepare_new_iteration(self):
        n_iter = self.sim_manager.n_iter

        with self.data_manager.lock:
            iter_group = self.data_manager.get_iter_group(n_iter)

            try:
                del iter_group['stringmethod']
            except KeyError:
                pass

            sm_global_group = self.data_manager.we_h5file.require_group('stringmethod')
            last_update = int(sm_global_group.attrs.get('last_update', 0))

        if n_iter - last_update < self.update_interval or n_iter < self.initial_update or not self.do_update:
            log.debug('Not updating string this iteration')
            return
        else:
            log.debug('Updating string - n_iter: {}'.format(n_iter))

        westpa.rc.pstatus('-westext.stringmethod -----------------\n')
        westpa.rc.pstatus('westext.stringmethod: Calculating average position in string images\n')
        westpa.rc.pflush()

        avg_pos, sum_bin_weight = self.get_avgpos(n_iter)

        westpa.rc.pstatus('westext.stringmethod: Updating string\n')
        westpa.rc.pflush()

        self.strings.update_string_centers(avg_pos, sum_bin_weight)

        westpa.rc.pstatus('westext.stringmethod: String lengths: {}\n'.format(self.strings.length))
        westpa.rc.pflush()

        # Update the bin definitions
        self.update_bin_mapper()

        sm_global_group.attrs['last_update'] = n_iter



import westpa
import itertools


def blocked_iter(blocksize, iterable, fillvalue=None):
    # From the Python "itertools recipes" (grouper)
    args = [iter(iterable)] * blocksize
    return itertools.zip_longest(fillvalue=fillvalue, *args)


class WESTPropagator:
    def __init__(self, rc=None):
        # For maximum flexibility, the basis states and initial states valid
        # at the point in the simulation when the propgator is used must be
        # available in several routines, and it is inconvenient to pass them
        # to every routine that needs them. A currently-reasonable-seeming solution
        # is to store at least the basis states and initial states necessary for
        # the current operation (propagation, etc). The set_basis_initial_states() function
        # accomplishes this. They are stored as dictionaries of state_id -> state,
        # so they can be looked up by ID without needing to store them all (and
        # thus potentially send them all over the wire when only one of them is needed, e.g.)
        self.basis_states = {}
        self.initial_states = {}

        self.rc = rc or westpa.rc

    def prepare_iteration(self, n_iter, segments):
        """Perform any necessary per-iteration preparation.  This is run by the work manager."""
        pass

    def finalize_iteration(self, n_iter, segments):
        """Perform any necessary post-iteration cleanup.  This is run by the work manager."""
        pass

    # Specific functions required by the WEST framework
    def get_pcoord(self, state):
        '''Get the progress coordinate of the given basis or initial state.'''
        raise NotImplementedError

    def gen_istate(self, basis_state, initial_state):
        '''Generate a new initial state from the given basis state.'''
        raise NotImplementedError

    def propagate(self, segments):
        """Propagate one or more segments, including any necessary per-iteration setup and teardown for this propagator."""
        raise NotImplementedError

    def clear_basis_initial_states(self):
        self.basis_states = {}
        self.initial_states = {}

    def update_basis_initial_states(self, basis_states, initial_states):
        self.basis_states.update({state.state_id: state for state in basis_states})
        self.initial_states.update({state.state_id: state for state in initial_states})



from h5py import File as H5File
import numpy as np
from westpa import rc
from westpa.tools import WESTParallelTool


class DurationCorrector(object):
    @staticmethod
    def from_kinetics_file(directh5, istate, fstate, dtau, n_iters=None):
        iter_slice = slice(n_iters)

        if isinstance(directh5, H5File):
            dataset = directh5['durations'][iter_slice]
        else:
            with H5File(directh5, 'r') as directh5:
                dataset = directh5['durations'][iter_slice]

        torf = np.logical_and(dataset['istate'] == istate, dataset['fstate'] == fstate)
        torf = np.logical_and(torf, dataset['weight'] > 0)

        durations = dataset['duration']
        weights = dataset['weight']

        weights[~torf] = 0.0  # mask off irrelevant flux

        return DurationCorrector(durations, weights, dtau)

    def __init__(self, durations, weights, dtau, maxduration=None):
        self.weights = np.array(weights)
        self.durations = np.array(durations)
        self.dtau = dtau
        self._f_tilde = None
        self._f_int1 = None

        if maxduration is None:
            self.maxduration = self.durations.shape[0]
        else:
            self.maxduration = maxduration

        if dtau is None:
            all_durations = []
            all_durations.extend(durations)
            all_durations.extend(np.arange(maxduration))
            uniq_durations = np.unique(all_durations)  # unique sorts automatically
            self.dtau = np.min(np.diff(uniq_durations))

        self._build_map()

    @property
    def event_duration_histogram(self):
        return self._f_tilde

    @property
    def cumulative_event_duration_histogram(self):
        return self._f_int1

    def _build_map(self):
        weights = self.weights
        durations = self.durations
        maxduration = self.maxduration
        dtau = self.dtau

        taugrid = np.arange(0, maxduration, dtau, dtype=float)
        f_map = np.zeros(weights.shape, dtype=int) - 1
        for i, tau in enumerate(taugrid):
            matches = np.logical_and(durations >= tau, durations < tau + dtau)
            f_map[matches] = i

        self.taugrid = taugrid
        self.f_map = f_map

    def correction(self, iters, freqs=None):
        r"""
        Return the correction factor

        __                              __  -1
        |  t=theta  tau=t                |
        |    |\      |\                  |
        |    |       | ~                 |
        |    |       | f(tau) dtau dt    |      * maxduration
        |   \|      \|                   |
        |   t=0    tau=0                 |
        |_                              _|

        where
            ~`                        ^
            f(tau) is proportional to f(tau)/(theta-tau), and is normalized to
                            ^
        integrate to 1, and f(tau) is sum of the weights of walkers with
        duration time tau.

        ---------
        Arguments
        ---------
        maxduration: the maximum duration time that could have been observed in
            the simulation, which is usually equal to the length of the
            simulation. This should be in units of tau.
        """

        if iters is None:
            iters = np.arange(len(self.weights))

        if freqs is None:
            freqs = np.ones(len(iters), dtype=float)

        maxduration = np.max(iters) + 1

        f_map = self.f_map[iters]
        weights = self.weights[iters]
        taugrid = self.taugrid  # [self.taugrid < maxduration]

        weights *= freqs[:, None]

        dtau = self.dtau

        f_tilde = np.zeros(len(taugrid), dtype=float)
        for i, tau in enumerate(taugrid):
            if tau < maxduration:
                f_tilde[i] = weights[f_map == i].sum() / (maxduration - tau + 1)

        if f_tilde.sum() != 0:
            f_tilde /= f_tilde.sum() * dtau

        self._f_tilde = f_tilde
        # now integrate f_tilde twice
        # integral1[t/dtau] gives the integral of f_tilde(tau) dtau from 0 to t
        self._f_int1 = integral1 = np.zeros(f_tilde.shape)

        for i, tau in enumerate(taugrid):
            if i > 0 and tau < maxduration:
                integral1[i] = np.trapz(f_tilde[: i + 1], taugrid[: i + 1])

        integral2 = np.trapz(integral1, taugrid)

        if integral2 == 0:
            return 0.0
        return maxduration / integral2


def get_raw_rates(directh5, istate, fstate, n_iters=None):
    rate_evol = directh5['rate_evolution'][slice(n_iters), istate, fstate]
    avg = rate_evol['expected']

    return avg


def calc_avg_rate(directh5_path, istate, fstate, **kwargs):
    """
    Return the raw or RED-corrected rate constant with the confidence interval.

    ---------
    Arguments
    ---------
    dt: timestep (ps)
    nstiter: duration of each iteration (number of steps)
    ntpr: report inteval (number of steps)

    """

    n_iters = kwargs.pop("n_iters", None)

    ntpr = kwargs.pop("report_interval", 20)
    nstiter = kwargs.pop("n_steps_iter", 1000)
    callback = kwargs.pop("callback", None)

    red = kwargs.pop("red", False)

    if len(kwargs) > 0:
        raise ValueError("unparsed kwargs")

    dtau = float(ntpr) / nstiter
    dc = None

    with H5File(directh5_path, 'r') as directh5:
        if n_iters is None:
            n_iters = directh5['rate_evolution'].shape[0]

        rate_evol = directh5['rate_evolution'][n_iters - 1, istate, fstate]
        rate = rate_evol['expected']

        if red:
            dc = DurationCorrector.from_kinetics_file(directh5, istate, fstate, dtau, n_iters)

    if callback is not None:
        kw = {"correction": dc}
        callback(**kw)

    iters = np.arange(n_iters)

    correction = dc.correction(iters) if dc else 1.0

    rate *= correction

    return rate


def calc_rates(directh5_path, istate, fstate, **kwargs):
    """
    Return the raw and RED-corrected rate constants vs. iterations.
    This code is faster than calling calc_rate() iteratively

    ---------
    Arguments
    ---------
    dt: timestep (ps)
    nstiter: duration of each iteration (number of steps)
    ntpr: report inteval (number of steps)

    """

    n_iters = kwargs.pop("n_iters", None)

    ntpr = kwargs.pop("report_interval", 20)
    nstiter = kwargs.pop("n_steps_iter", 1000)
    callback = kwargs.pop("callback", None)

    red = kwargs.pop("red", False)

    if len(kwargs) > 0:
        raise ValueError("unparsed kwargs")

    dtau = float(ntpr) / nstiter
    dc = None

    with H5File(directh5_path, 'r') as directh5:
        rate_evol, cilb, ciub = get_raw_rates(directh5, istate, fstate, n_iters)
        if n_iters is None:
            n_iters = len(rate_evol)
        if red:
            dc = DurationCorrector.from_kinetics_file(directh5, istate, fstate, dtau, n_iters)

    if callback is not None:
        kw = {"correction": dc}
        callback(**kw)

    raw_rates = np.zeros(n_iters)

    rates = np.zeros(n_iters)

    for i in range(n_iters):
        i_iter = i + 1
        print("\riter %d/%d (%3.0f%%)" % (i_iter, n_iters, i_iter * 100.0 / n_iters), end="")

        r = rate_evol[i]

        iters = np.arange(i_iter)

        correction = dc.correction(iters) if dc else 1.0

        raw_rates[i] = r
        rates[i] = raw_rates[i] * correction

    print("\n")

    return rates


class RateCalculator:
    def __init__(self, directh5, istate, fstate, assignh5=None, **kwargs):
        n_iters = kwargs.pop("n_iters", None)
        ntpr = kwargs.pop("report_interval", 20)
        nstiter = kwargs.pop("n_steps_iter", 1000)

        if len(kwargs) > 0:
            for k in kwargs:
                print(k)
            raise ValueError("unparsed kwargs")

        dtau = float(ntpr) / nstiter

        with H5File(directh5, 'r') as f:
            state_labels = {}
            for i, raw_label in enumerate(f['state_labels']):
                label = raw_label.decode() if isinstance(raw_label, bytes) else raw_label
                state_labels[label] = i
            if istate not in state_labels:
                raise ValueError(f"istate not found: {istate}, available options are {list(state_labels.keys())}")
            if fstate not in state_labels:
                raise ValueError(f"istate not found: {fstate}, available options are {list(state_labels.keys())}")
            istate = state_labels[istate]
            fstate = state_labels[fstate]
            cond_fluxes = f['conditional_fluxes'][slice(n_iters), istate, fstate]

        if assignh5 is not None:
            with H5File(assignh5, 'r') as f:
                pops = f['labeled_populations'][slice(n_iters)]
                pops = pops.sum(axis=2)
        else:
            pops = None

        self._dc = None
        self._pops = pops
        self._cond_fluxes = cond_fluxes
        self._dtau = dtau
        self._directh5 = directh5
        self._assignh5 = assignh5
        self._istate = istate
        self._fstate = fstate

    @property
    def conditional_fluxes(self):
        return self._cond_fluxes

    @property
    def populations(self):
        return self._pops

    @property
    def tau(self):
        return self._tau

    @property
    def dtau(self):
        return self._dtau

    @property
    def istate(self):
        return self._istate

    @property
    def fstate(self):
        return self._fstate

    @property
    def n_iters(self):
        return len(self.conditional_fluxes)

    def _get_corrector(self):
        if self._dc is None:
            with H5File(self._directh5, 'r') as f:
                self._dc = DurationCorrector.from_kinetics_file(f, self.istate, self.fstate, self.dtau, self.n_iters)

        return self._dc

    def calc_rate(self, i_iter=None, red=False, **kwargs):
        if i_iter is None:
            i_iter = self.n_iters

        dc = self._get_corrector() if red else None
        found = False
        with H5File(self._directh5, 'r') as f:
            for i in range(f['rate_evolution'].shape[0]):
                rate_evol = f['rate_evolution'][i, self.istate, self.fstate]
                start = rate_evol['iter_start']
                stop = rate_evol['iter_stop']

                if i_iter >= start and i_iter < stop:
                    rate = rate_evol['expected']
                    found = True
                    break

            if not found:
                self.log.error("Can't find rate evolution data for iteration %d!" % i_iter)

        if dc:
            iters = np.arange(i_iter)
            correction = dc.correction(iters)
            rate *= correction

        return rate

    def calc_rates(self, n_iters=None, **kwargs):
        if n_iters is None:
            n_iters = self.n_iters

        rates = np.zeros(n_iters)

        for i in range(n_iters):
            i_iter = i + 1
            print("\riter %d/%d (%3.0f%%)" % (i_iter, n_iters, i_iter * 100.0 / n_iters), end="")

            r = self.calc_rate(i_iter, **kwargs)

            rates[i] = r

        print("\n")

        return rates


class WRed(WESTParallelTool):
    prog = 'w_red'
    description = '''\
Apply the RED scheme to estimate steady-state WE fluxes from
shorter trajectories.

-----------------------------------------------------------------------------
Source data
-----------------------------------------------------------------------------

Source data is provided as a w_ipa "scheme" which is typically defined
in the west.cfg file. For instance, if a user wishes to estimate RED
fluxes for a scheme named "DEFAULT" that argument would be provided
to w_red and WRed would estimate RED fluxes based off of the data
contained in the assign.h5 and direct.h5 files in ANALYSIS/DEFAULT.

'''

    def __init__(self):
        super().__init__()

    def go(self):
        try:
            rc.config['west']['analysis']['red']
        except Exception:
            raise ValueError('No RED parameters are specified in west.cfg.')
        try:
            rc.config['west']['analysis']['red']['scheme']
        except Exception:
            raise ValueError('No scheme specified for RED calculation in west.cfg.')
        try:
            rc.config['west']['analysis']['red']['istate_label']
        except Exception:
            raise ValueError('No intial state label specified for RED calculation in west.cfg.')
        try:
            rc.config['west']['analysis']['red']['fstate_label']
        except Exception:
            raise ValueError('No final state label specified for RED calculation in west.cfg.')
        try:
            rc.config['west']['analysis']['red']['nstiter']
        except Exception:
            raise ValueError('Time step not specified in west.cfg.')
        try:
            rc.config['west']['analysis']['red']['nstrep']
        except Exception:
            raise ValueError('Time step not specified in west.cfg.')

        if rc.config['west']['analysis']['kinetics']['evolution'] == "cumulative":
            pass
        else:
            print("Only RED estimates with cumulative averaging are supported at this time.")
            exit()

        config = rc.config
        adir = config.get(['west', 'analysis', 'directory'])
        name = config.get(['west', 'analysis', 'red', 'scheme'])
        istate = config.get(['west', 'analysis', 'red', 'istate_label'])
        fstate = config.get(['west', 'analysis', 'red', 'fstate_label'])
        n_steps_per_iter = config.get(['west', 'analysis', 'red', 'nstiter'])
        n_steps_per_report = config.get(['west', 'analysis', 'red', 'nstrep'])

        directh5path = '%s/%s/direct.h5' % (adir, name)
        assignh5path = '%s/%s/assign.h5' % (adir, name)

        print('\nConfig successfully read from west.cfg:')
        print('--------------------------------------')
        print('scheme: %s' % name)
        print('states: %s -> %s' % (istate, fstate))
        print('nstiter: %s' % n_steps_per_iter)
        print('nstrep: %s' % n_steps_per_report)
        print('--------------------------------------')
        print('\nEstimating RED fluxes...')
        rater = RateCalculator(
            directh5path,
            istate,
            fstate,
            n_steps_iter=n_steps_per_iter,
            report_interval=n_steps_per_report,
            assignh5=assignh5path,
        )

        rates = rater.calc_rates(red=True, callback=None)

        with H5File(directh5path, "r+") as dest_file:
            try:
                dest_file.create_dataset('red_flux_evolution', data=rates)
                print('saved RED fluxes to red_flux_evolution in ANALYSIS/%s/direct.h5' % name)
            except Exception:
                warning = input('Dataset already exists! Overwrite? (y/n)')
                if warning == "y":
                    dest_file['red_flux_evolution'][...] = rates
                    print('saved RED fluxes to red_flux_evolution in ANALYSIS/%s/direct.h5' % name)
                elif warning == "n":
                    np.save('ANALYSIS/%s/red.npy' % name, rates)
                    print('saved RED fluxes to red_flux_evolution.npy in ANALYSIS/%s' % name)
                else:
                    print('red rates not saved. Exiting...')
                    exit


def entry_point():
    WRed().main()


if __name__ == '__main__':
    entry_point()



import argparse
import logging

import numpy as np

import westpa

from oldtools.aframe import (
    WESTAnalysisTool,
    BinningMixin,
    WESTDataReaderMixin,
    IterRangeMixin,
    MCBSMixin,
    TransitionAnalysisMixin,
    KineticsAnalysisMixin,
    CommonOutputMixin,
    BFDataManager,
    BFTransitionAnalysisMixin,
)

log = logging.getLogger('w_ttimes')


ciinfo_dtype = np.dtype([('expectation', np.float64), ('ci_lower', np.float64), ('ci_upper', np.float64)])


class WTTimesBase:
    def __init__(self):
        super().__init__()

        self.ed_stats_filename = None
        self.fpt_stats_filename = None
        self.flux_stats_filename = None
        self.rate_stats_filename = None
        self.suppress_headers = None
        self.print_bin_labels = None

        self.ttimes_group = None

        self.durations = None
        self.fpts = None
        self.fluxes = None
        self.rates = None

    def add_args(self, parser, upcall=True):
        '''Add arguments to a parser common to all analyses of this type.'''
        if upcall:
            try:
                upfunc = super().add_args
            except AttributeError:
                pass
            else:
                upfunc(parser)

        output_options = parser.add_argument_group('kinetics analysis output options')
        output_options.add_argument(
            '--edstats',
            dest='ed_stats',
            default='edstats.txt',
            help='Store event duration statistics in ED_STATS (default: edstats.txt)',
        )
        if self.bf_mode:
            output_options.add_argument(
                '--fptstats',
                dest='fpt_stats',
                default='fptstats.txt',
                help='Store first passage time statistics in FPT_STATS (default: fptstats.txt).',
            )
        else:
            output_options.add_argument(
                '--fptstats', dest='fpt_stats', help='Store first passage time statistics in FPT_STATS (default: do not store).'
            )
        output_options.add_argument(
            '--fluxstats',
            dest='flux_stats',
            default='fluxstats.txt',
            help='Store flux statistics in FLUX_STATS (default: fluxstats.txt)',
        )
        output_options.add_argument(
            '--ratestats',
            dest='rate_stats',
            default='ratestats.txt',
            help='Store rate statistics in RATE_STATS (default: ratestats.txt)',
        )
        self.add_common_output_args(output_options)

    def process_args(self, args, upcall=True):
        self.ed_stats_filename = args.ed_stats
        self.fpt_stats_filename = args.fpt_stats
        self.flux_stats_filename = args.flux_stats
        self.rate_stats_filename = args.rate_stats
        self.process_common_output_args(args)
        self.calc_fpts = bool(args.fpt_stats)

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def gen_stats(self):
        self.require_transitions_group()
        westpa.rc.pstatus('Analyzing transition statistics...')

        dt = self.dt
        n_sets = self.mcbs_nsets
        lbi, ubi = self.calc_ci_bound_indices()

        total_time = self.get_total_time()
        transdat_ds = self.trans_h5group['transitions']
        transdat_ibin = transdat_ds['initial_bin']

        if not self.bf_mode:
            transdat_niter = transdat_ds['n_iter']
            transdat_in_range = (transdat_niter >= self.first_iter) & (transdat_niter <= self.last_iter)

        durations = np.zeros((self.n_bins, self.n_bins), ciinfo_dtype)
        fpts = np.zeros((self.n_bins, self.n_bins), ciinfo_dtype)
        fluxes = np.zeros((self.n_bins, self.n_bins), ciinfo_dtype)
        rates = np.zeros((self.n_bins, self.n_bins), ciinfo_dtype)

        syn_avg_durations = np.empty((n_sets,), np.float64)
        syn_avg_fpts = np.empty((n_sets,), np.float64)
        syn_avg_fluxes = np.empty((n_sets,), np.float64)
        syn_avg_rates = np.empty((n_sets,), np.float64)

        w_n_bins = len(str(self.n_bins))
        w_n_sets = len(str(n_sets))

        for ibin in self.analysis_initial_bins:
            if self.bf_mode:
                trans_ibin = transdat_ds[transdat_ibin == ibin]
            else:
                trans_ibin = transdat_ds[(transdat_ibin == ibin) & transdat_in_range]

            for fbin in self.analysis_final_bins:
                # trans_ifbins = trans_ibin[trans_ibin['final_bin'] == fbin]
                trans_ifbins = np.extract(trans_ibin['final_bin'] == fbin, trans_ibin)
                dlen = len(trans_ifbins)

                if not dlen:
                    continue

                trans_weights = trans_ifbins['final_weight']
                trans_durations = trans_ifbins['duration']
                trans_fpts = trans_ifbins['fpt']
                trans_ibinprobs = trans_ifbins['initial_bin_pop']

                durations[ibin, fbin]['expectation'] = np.average(trans_durations, weights=trans_weights) * dt
                fpts[ibin, fbin]['expectation'] = np.average(trans_fpts, weights=trans_weights) * dt
                avg_flux = trans_weights.sum() / total_time
                fluxes[ibin, fbin]['expectation'] = avg_flux
                rates[ibin, fbin]['expectation'] = avg_flux / trans_ibinprobs.mean()

                for iset in range(n_sets):
                    westpa.rc.pstatus(
                        '\r  {:{w_n_bins}d}->{:<{w_n_bins}d} set {:{w_n_sets}d}/{:<{w_n_sets}d}, set size {:<20d}'.format(
                            ibin, fbin, iset + 1, n_sets, dlen, w_n_bins=w_n_bins, w_n_sets=w_n_sets
                        ),
                        end='',
                    )
                    westpa.rc.pflush()
                    indices = np.random.randint(dlen, size=(dlen,))
                    # syn_weights   = trans_weights[indices]
                    # syn_durations = trans_durations[indices]
                    # syn_fpts      = trans_fpts[indices]
                    # syn_ibinprobs = trans_ibinprobs[indices]
                    syn_weights = trans_weights.take(indices)
                    syn_durations = trans_durations.take(indices)
                    syn_fpts = trans_fpts.take(indices)
                    syn_ibinprobs = trans_ibinprobs.take(indices)

                    syn_avg_durations[iset] = np.average(syn_durations, weights=syn_weights) * dt
                    syn_avg_fpts[iset] = np.average(syn_fpts, weights=syn_weights) * dt
                    syn_avg_fluxes[iset] = syn_weights.sum() / total_time
                    syn_avg_rates[iset] = syn_avg_fluxes[iset] / syn_ibinprobs.mean()

                    del indices, syn_weights, syn_durations, syn_ibinprobs, syn_fpts

                syn_avg_durations.sort()
                syn_avg_fpts.sort()
                syn_avg_fluxes.sort()
                syn_avg_rates.sort()

                durations[ibin, fbin]['ci_lower'] = syn_avg_durations[lbi]
                durations[ibin, fbin]['ci_upper'] = syn_avg_durations[ubi]

                fpts[ibin, fbin]['ci_lower'] = syn_avg_fpts[lbi]
                fpts[ibin, fbin]['ci_upper'] = syn_avg_fpts[ubi]

                fluxes[ibin, fbin]['ci_lower'] = syn_avg_fluxes[lbi]
                fluxes[ibin, fbin]['ci_upper'] = syn_avg_fluxes[ubi]

                rates[ibin, fbin]['ci_lower'] = syn_avg_rates[lbi]
                rates[ibin, fbin]['ci_upper'] = syn_avg_rates[ubi]

                del trans_weights, trans_durations, trans_ibinprobs, trans_ifbins, trans_fpts
            westpa.rc.pstatus()
            del trans_ibin

        for dsname, data in (('duration', durations), ('fpt', fpts), ('flux', fluxes), ('rate', rates)):
            try:
                del self.ttimes_group[dsname]
            except KeyError:
                pass

            ds = self.ttimes_group.create_dataset(dsname, data=data)
            attrs = ds.attrs
            attrs['dt'] = dt
            attrs['total_time'] = total_time
            attrs['ci_alpha'] = self.mcbs_alpha
            attrs['ci_n_sets'] = self.mcbs_nsets

            if not self.bf_mode:
                self.record_data_iter_range(ds)
            self.record_data_binhash(ds)

        attrs = self.ttimes_group.attrs
        attrs = ds.attrs
        attrs['dt'] = dt
        attrs['total_time'] = total_time
        attrs['ci_alpha'] = self.mcbs_alpha
        attrs['ci_n_sets'] = self.mcbs_nsets

        self.durations = durations
        self.fluxes = fluxes
        self.rates = rates

        if not self.bf_mode:
            self.record_data_iter_range(self.ttimes_group)
        self.record_data_binhash(self.ttimes_group)

    def summarize_stats(self):
        for array, dsname, argname, title in (
            (self.durations, 'duration', 'ed_stats_filename', 'event duration'),
            (self.fpts, 'fpt', 'fpt_stats_filename', 'first passage time'),
            (self.fluxes, 'flux', 'flux_stats_filename', 'flux'),
            (self.rates, 'rate', 'rate_stats_filename', 'rate'),
        ):
            filename = getattr(self, argname)
            if filename:
                if array is None:
                    try:
                        array = self.ttimes_group[dsname]
                    except KeyError:
                        westpa.rc.pstatus('{} data not found in {}'.format(title, self.anal_h5name))
                        continue

                self.summarize_ci(
                    filename,
                    array,
                    title,
                    self.mcbs_display_confidence,
                    headers=(not self.suppress_headers),
                    labels=self.print_bin_labels,
                )

    def summarize_ci(self, filename, array, title, confidence, headers, labels):
        format_2d = (
            '{ibin:{mw}d}    {fbin:{mw}d}    {0:20.15g}    {1:20.15g}    {2:20.15g}    {3:20.15g}    {4:20.15g}    {5:20.15g}\n'
        )
        max_ibin_width = len(str(self.n_bins - 1))

        outfile = open(filename, 'wt')
        if headers:
            outfile.write(
                '''\
# {title:} statistics
# confidence interval = {confidence}%
# ----
# column 0: initial bin index
# column 1: final bin index
# column 2: lower bound of confidence interval
# column 3: upper bound of confidence interval
# column 4: width of confidence interval
# column 5: relative width of confidence interval [abs(width/average)]
# column 6: symmetrized error [max(upper-average, average-lower)]
# ----
'''.format(
                    title=title, confidence=confidence
                )
            )
            if labels:
                self.write_bin_labels(outfile)
                outfile.write('----\n')

        for ibin in self.analysis_initial_bins:
            for fbin in self.analysis_final_bins:
                mean = array[ibin, fbin]['expectation']
                lb = array[ibin, fbin]['ci_lower']
                ub = array[ibin, fbin]['ci_upper']
                ciwidth = ub - lb
                relciwidth = abs(ciwidth / mean)
                symmerr = max(mean - lb, ub - mean)

                outfile.write(
                    format_2d.format(
                        *list(map(float, (mean, lb, ub, ciwidth, relciwidth, symmerr))), ibin=ibin, fbin=fbin, mw=max_ibin_width
                    )
                )

    def main(self):
        parser = argparse.ArgumentParser('w_ttimes', description=self.description)
        westpa.rc.add_args(parser)
        self.add_args(parser)

        args = parser.parse_args()
        westpa.rc.process_args(args, config_required=False)
        self.process_args(args)

        self.check_iter_range()
        self.check_bin_selection()
        self.open_analysis_backing()
        self.ttimes_group = self.require_analysis_group('w_ttimes', replace=False)
        self.require_bin_assignments()
        self.require_transitions()
        self.gen_stats()
        self.summarize_stats()


class WTTimesWE(
    WTTimesBase,
    CommonOutputMixin,
    MCBSMixin,
    KineticsAnalysisMixin,
    TransitionAnalysisMixin,
    BinningMixin,
    IterRangeMixin,
    WESTDataReaderMixin,
    WESTAnalysisTool,
):
    description = 'Trace the WEST trajectory tree and report on transition kinetics.'

    def __init__(self):
        super().__init__()


class WTTimesBF(
    WTTimesBase, CommonOutputMixin, MCBSMixin, KineticsAnalysisMixin, BFTransitionAnalysisMixin, BFDataManager, WESTAnalysisTool
):
    description = 'Trace one or more brute force trajectories and report on transition kinetics.'
    default_chunksize = 65536 * 4

    def __init__(self):
        super().__init__()
        self.bf_mode = True
        self.config_required = False
        self.usecols = None
        self.input_files = None

    def check_iter_range(self):
        pass  # do nothing, since we don't do iteration ranges for brute force

    def get_total_time(self):
        self.require_bf_h5file()
        return np.add.reduce([self.get_traj_len(traj_id) - 1 for traj_id in range(self.get_n_trajs())]) * self.dt



"""WEST run control and configuration routines"""

import math
import os
import sys
import errno
import logging
import warnings
from copy import deepcopy
from datetime import timedelta

import numpy as np

import westpa
import westpa.core.data_manager
from westpa.core.binning.assign import BinMapper
from westpa.core.binning import RectilinearBinMapper, RecursiveBinMapper, MABBinMapper, BinlessMapper
from .yamlcfg import YAMLConfig
from .yamlcfg import YAMLSystem
from . import extloader
from ..work_managers import SerialWorkManager

log = logging.getLogger('westpa.rc')


def bins_from_yaml_dict(bin_dict):
    kwargs = deepcopy(bin_dict)
    typename = kwargs.pop('type')

    try:
        mapper_type = getattr(sys.modules['westpa.core.binning'], typename)
    except AttributeError:
        try:
            mapper_type = extloader.get_object(typename)
        except AttributeError:
            raise KeyError('unknown bin mapper type {!r}'.format(typename))

    if not issubclass(mapper_type, BinMapper):
        raise ValueError('{} is not a BinMapper'.format(mapper_type.__name__))

    if mapper_type is RectilinearBinMapper:
        boundary_lists = kwargs.pop('boundaries', None)
        if boundary_lists is None:
            raise KeyError('RectilinearBinMapper: missing boundaries')
        parsed_lists = boundary_lists[:]
        for iboundary, boundary in enumerate(boundary_lists):
            if boundary.__class__ == str:
                parsed_lists[iboundary] = parsePCV(boundary)[0]
            else:
                parsed_lists[iboundary] = list(map((lambda x: float(x) if isinstance(x, str) else x), boundary))
        return RectilinearBinMapper(parsed_lists)
    elif mapper_type is RecursiveBinMapper:
        base_mapper_config = kwargs.pop('base', None)
        base_mapper_config = kwargs.pop('base_mapper', base_mapper_config)
        if base_mapper_config is None:
            raise KeyError('RecursiveBinMapper: missing base_mapper')

        base_mapper = bins_from_yaml_dict(base_mapper_config)
        start_index = kwargs.pop('start_index', 0)

        rec_mapper = RecursiveBinMapper(base_mapper, start_index)
        mapper_configs = kwargs.pop('mappers')
        mappers = []
        if mapper_configs is not None:
            for config in mapper_configs:
                replaced_bin = config.pop('replaces_bin_at', None)
                replaced_bin = config.pop('at', replaced_bin)
                if replaced_bin is None:
                    raise KeyError('RecursiveBinMapper: missing replaces_bin_at for ' 'at least one of the child mappers')
                mapper = bins_from_yaml_dict(config)
                mappers.append((mapper, replaced_bin))

        for mapper, replaced_bin in mappers:
            rec_mapper.add_mapper(mapper, replaced_bin)

        return rec_mapper
    else:
        try:
            return mapper_type(**kwargs)
        except Exception:
            log.exception('exception instantiating mapper')
            raise


def detect_mab_mapper(mapper):
    if isinstance(mapper, MABBinMapper):
        return True
    elif isinstance(mapper, RecursiveBinMapper):
        if detect_mab_mapper(mapper.base_mapper):
            return True

        for ibin in mapper._recursion_targets:
            rec_mapper = mapper._recursion_targets[ibin]
            if detect_mab_mapper(rec_mapper):
                return True
    else:
        return False


def detect_binless_mapper(mapper):
    if isinstance(mapper, BinlessMapper):
        return True
    elif isinstance(mapper, RecursiveBinMapper):
        if detect_binless_mapper(mapper.base_mapper):
            return True

        for ibin in mapper._recursion_targets:
            rec_mapper = mapper._recursion_targets[ibin]
            if detect_binless_mapper(rec_mapper):
                return True
    else:
        return False


def parsePCV(pc_str):
    # Execute arbitrary code within a limited
    # scope to avoid nastyness. Stolen fully from
    # other parts of the WESTPA code.
    namespace = {'math': math, 'numpy': np, 'np': np, 'inf': float('inf')}

    arr = np.array(eval(pc_str, namespace))
    if arr.ndim == 0:
        arr.shape = (1, 1)
    elif arr.ndim == 1:
        arr.shape = (1,) + arr.shape
    else:
        raise ValueError('too many dimensions')
    # return list(arr[...])
    return arr[...]


def lazy_loaded(backing_name, loader, docstring=None):
    def getter(self):
        obj = getattr(self, backing_name, None)
        if obj is None:
            obj = loader()
            setattr(self, backing_name, obj)
        return obj

    def setter(self, val):
        setattr(self, backing_name, val)

    def deleter(self):
        delattr(self, backing_name)
        setattr(self, backing_name, None)

    return property(getter, setter, deleter, docstring)


class WESTRC:
    '''A class, an instance of which is accessible as ``westpa.rc``, to handle global issues for WEST-PA code,
    such as loading modules and plugins, writing output based on verbosity level, adding default command line options,
    and so on.'''

    # Runtime config file management
    ENV_RUNTIME_CONFIG = 'WESTRC'
    RC_DEFAULT_FILENAME = 'west.cfg'

    def __init__(self):
        self.verbosity = None
        self.rcfile = os.environ.get(self.ENV_RUNTIME_CONFIG) or self.RC_DEFAULT_FILENAME

        self.config = YAMLConfig()
        try:
            self.process_name = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        except (TypeError, IndexError):
            self.process_name = "unknown"

        # Crucial simulation and analysis drivers
        self._system = None
        self._data_manager = None
        self._sim_manager = None
        self._we_driver = None
        self._propagator = None

        self.work_manager = SerialWorkManager()

        self.status_stream = sys.stdout

    def add_args(self, parser):
        group = parser.add_argument_group('general options')
        group.add_argument(
            '-r',
            '--rcfile',
            metavar='RCFILE',
            dest='rcfile',
            default=(os.environ.get(self.ENV_RUNTIME_CONFIG) or self.RC_DEFAULT_FILENAME),
            help='use RCFILE as the WEST run-time configuration file (default: %(default)s)',
        )

        egroup = group.add_mutually_exclusive_group()
        egroup.add_argument(
            '--quiet', dest='verbosity', action='store_const', const='quiet', help='emit only essential information'
        )
        egroup.add_argument('--verbose', dest='verbosity', action='store_const', const='verbose', help='emit extra information')
        egroup.add_argument(
            '--debug',
            dest='verbosity',
            action='store_const',
            const='debug',
            help='enable extra checks and emit copious information',
        )

        group.add_argument('--version', action='version', version='WEST version %s' % westpa.__version__)

    @property
    def verbose_mode(self):
        return self.verbosity in ('verbose', 'debug')

    @property
    def debug_mode(self):
        return self.verbosity == 'debug'

    @property
    def quiet_mode(self):
        return self.verbosity == 'quiet'

    def process_args(self, args, config_required=True):
        self.cmdline_args = args
        self.verbosity = args.verbosity

        if args.rcfile:
            self.rcfile = args.rcfile

        try:
            self.read_config()
        except IOError as e:
            if e.errno == errno.ENOENT and not config_required:
                pass
            else:
                raise
        self.config_logging()
        self.config['args'] = {k: v for k, v in args.__dict__.items() if not k.startswith('_')}
        self.process_config()
        self.time_config()

    def time_config(self):
        '''Convert non-pyYAML accepted time formats into seconds'''
        max_run_wallclock = self.config.get(['west', 'propagation', 'max_run_wallclock'], None)
        if isinstance(max_run_wallclock, str):
            time_formats = ['days', 'hours', 'minutes', 'seconds']
            wallclock_str = max_run_wallclock.split(':')
            kwargs = {key: int(val) for key, val in zip(time_formats[-len(wallclock_str) :], wallclock_str)}
            try:
                diff = int(timedelta(**kwargs).total_seconds())
                self.config['west']['propagation']['max_run_wallclock'] = diff
                log.debug(f'Automatically converted {max_run_wallclock=} to {diff}')
            except ValueError:
                log.debug(f'Failed to convert {max_run_wallclock=} to seconds.')

    def process_config(self):
        log.debug('config: {!r}'.format(self.config))
        sys.path.extend(self.config.get_pathlist(['west', 'drivers', 'module_path'], []))
        try:
            sys.path.append(os.environ['WEST_SIM_ROOT'])
        except KeyError:
            pass

    def read_config(self, filename=None):
        if filename:
            self.rcfile = filename

        if 'WEST_SIM_ROOT' not in os.environ:
            # sys.stderr.write('-- WARNING -- setting $WEST_SIM_ROOT to current directory ({})\n'.format(os.getcwd()))
            os.environ['WEST_SIM_ROOT'] = os.getcwd()

        self.config.update_from_file(self.rcfile)

    def config_logging(self):
        import logging.config

        logging_config = {
            'version': 1,
            'incremental': False,
            'formatters': {
                'standard': {'format': '-- %(levelname)-8s [%(name)s] -- %(message)s'},
                'debug': {
                    'format': '''\
-- %(levelname)-8s %(asctime)24s PID %(process)-12d TID %(thread)-20d
   from logger "%(name)s"
   at location %(pathname)s:%(lineno)d [%(funcName)s()]
   ::
   %(message)s
'''
                },
            },
            'handlers': {'console': {'class': 'logging.StreamHandler', 'stream': 'ext://sys.stdout', 'formatter': 'standard'}},
            'loggers': {
                'west': {'handlers': ['console'], 'propagate': False},
                'westpa': {'handlers': ['console'], 'propagate': False},
                'oldtools': {'handlers': ['console'], 'propagate': False},
                'westtools': {'handlers': ['console'], 'propagate': False},
                'westext': {'handlers': ['console'], 'propagate': False},
                'work_managers': {'handlers': ['console'], 'propagate': False},
                'py.warnings': {'handlers': ['console'], 'propagate': False},
            },
            'root': {'handlers': ['console']},
        }

        logging_config['loggers'][self.process_name] = {'handlers': ['console'], 'propagate': False}

        if self.verbosity == 'debug':
            logging_config['root']['level'] = 5  # 'DEBUG'
            logging_config['handlers']['console']['formatter'] = 'debug'
        elif self.verbosity == 'verbose':
            logging_config['root']['level'] = 'INFO'
        else:
            logging_config['root']['level'] = 'WARNING'

        logging.config.dictConfig(logging_config)
        logging_config['incremental'] = True

        if self.verbosity == 'debug':
            warnings.resetwarnings()
            warnings.simplefilter("default")
            logging.captureWarnings(True)
        else:
            if not sys.warnoptions:
                warnings.simplefilter("ignore")
            logging.captureWarnings(False)

    def pstatus(self, *args, **kwargs):
        fileobj = kwargs.pop('file', self.status_stream)
        if kwargs.get('termonly', False) and not fileobj.isatty():
            return
        if self.verbosity != 'quiet':
            print(*args, file=fileobj, **kwargs)

    def pstatus_term(self, *args, **kwargs):
        fileobj = kwargs.pop('file', self.status_stream)
        if fileobj.isatty() and self.verbosity != 'quiet':
            print(*args, file=fileobj, **kwargs)

    def pflush(self):
        for stream in (self.status_stream, sys.stdout, sys.stderr):
            try:
                stream.flush()
            except AttributeError:
                pass

    def detect_mab_mapper(self):
        bin_dict = self.config.get(['west', 'system', 'system_options', 'bins'])
        use_mab = False
        if bin_dict is not None:
            mapper = bins_from_yaml_dict(bin_dict)
            use_mab = detect_mab_mapper(mapper)

        return use_mab

    def detect_binless_mapper(self):
        bin_dict = self.config.get(['west', 'system', 'system_options', 'bins'])
        use_binless = False
        if bin_dict is not None:
            mapper = bins_from_yaml_dict(bin_dict)
            use_binless = detect_binless_mapper(mapper)

        return use_binless

    def new_sim_manager(self):
        drivername = self.config.get(['west', 'drivers', 'sim_manager'], 'default')

        if drivername.lower() == 'default':
            use_mab = self.detect_mab_mapper()
            use_binless = self.detect_binless_mapper()

            if use_mab:
                from .binning.mab_manager import MABSimManager

                sim_manager = MABSimManager(rc=self)
            elif use_binless:
                from .binning.binless_manager import BinlessSimManager

                sim_manager = BinlessSimManager(rc=self)
            else:
                from .sim_manager import WESimManager

                sim_manager = WESimManager(rc=self)
        else:
            sim_manager = extloader.get_object(drivername)(rc=self)

        log.debug('loaded simulation manager {!r}'.format(sim_manager))
        return sim_manager

    def get_sim_manager(self):
        if self._sim_manager is None:
            self._sim_manager = self.new_sim_manager()
        return self._sim_manager

    def new_data_manager(self):
        import westpa

        drivername = self.config.get(['west', 'drivers', 'data_manager'], 'hdf5')
        if drivername.lower() in ('hdf5', 'default'):
            data_manager = westpa.core.data_manager.WESTDataManager()
        else:
            data_manager = extloader.get_object(drivername)(rc=self)
        log.debug('loaded data manager: {!r}'.format(data_manager))
        return data_manager

    def get_data_manager(self):
        if self._data_manager is None:
            self._data_manager = self.new_data_manager()
        return self._data_manager

    def new_we_driver(self):
        import westpa

        drivername = self.config.get(['west', 'drivers', 'we_driver'], 'default')

        if drivername.lower() == 'default':
            use_mab = self.detect_mab_mapper()
            use_binless = self.detect_binless_mapper()
            if use_mab:
                from .binning.mab_driver import MABDriver

                we_driver = MABDriver()
            elif use_binless:
                from .binning.binless_driver import BinlessDriver

                we_driver = BinlessDriver()
            else:
                from .we_driver import WEDriver

                we_driver = WEDriver()
        else:
            we_driver = extloader.get_object(drivername)(rc=self)

        log.debug('loaded WE algorithm driver: {!r}'.format(we_driver))

        subgroup_function = self.config.get(['west', 'drivers', 'subgroup_function'], 'default')
        if subgroup_function.lower() == 'default':
            try:
                subgroup_function = 'westpa.core.we_driver._group_walkers_identity'
                we_driver.subgroup_function = westpa.core.we_driver._group_walkers_identity
            except Exception:
                pass
        else:
            we_driver.subgroup_function = extloader.get_object(subgroup_function)
        we_driver.subgroup_function_kwargs = self.config.get(['west', 'drivers', 'subgroup_arguments'])
        # Necessary if the user hasn't specified any options.
        if we_driver.subgroup_function_kwargs is None:
            we_driver.subgroup_function_kwargs = {}
        log.debug('loaded WE algorithm driver subgrouping function {!r}'.format(subgroup_function))
        log.debug('WE algorithm driver subgrouping function kwargs: {!r}'.format(we_driver.subgroup_function_kwargs))

        return we_driver

    def get_we_driver(self):
        if self._we_driver is None:
            self._we_driver = self.new_we_driver()
        return self._we_driver

    def new_propagator(self):
        drivername = self.config.require(['west', 'propagation', 'propagator'])
        if drivername.lower() == 'executable':
            from westpa.core.propagators.executable import ExecutablePropagator

            propagator = ExecutablePropagator()
        else:
            propagator = extloader.get_object(drivername)(rc=self)
        log.debug('loaded propagator {!r}'.format(propagator))
        return propagator

    def get_propagator(self):
        if self._propagator is None:
            self._propagator = self.new_propagator()
        return self._propagator

    def new_system_driver(self):
        '''
        Returns a new system object either from the driver OR from the YAML
        file. Currently builds off of system then updates with YAML
        overwriting the previous settings if both are specified.

        There are no default settings, all settings MUST be specified
        in the config YAML file OR the system driver.

        Settings that are specified here are:
          Progress coordinate settings:
            Progress coordinate dimensionality
            Progress coordinate length/number of data points in each tau
            Progress coordinate data type
          Bin settings:
            Bin mapper type
            Bins
            Target simulation counts for each bin
          Generic setting for flexibility:
            Both paths allow for generic attribute setting for
            future flexibility.
        '''

        # First step is to see if we have a driver specified
        # if so load that one first, setting the defaults for
        # YAML to possibly modify.

        # I will keep this as is for now since I like the idea
        # of being able to set some defaults from the system
        # driver and then just modify the YAML instead for basic
        # stuff. This might make it easier to set up systems since
        # playing around with the YAML format is easier.

        # Still still has the end-user issue of possibly confusing
        # people, maybe KISS is better, not sure. I'll keep this
        # for development purposes if nothing else.

        system = None
        # Get method checks for us
        sysdrivername = self.config.get(['west', 'system', 'driver'])
        if not sysdrivername:
            # Warn user that driver is not specified
            log.info("System driver not specified")
        else:
            log.info('loading system driver %r' % sysdrivername)
            system = extloader.get_object(sysdrivername)(rc=self)
            log.debug('loaded system driver {!r}'.format(system))
            system.initialize()
        # Second let's see if we have info in the YAML file
        yamloptions = self.config.get(['west', 'system', 'system_options'])
        if not yamloptions:
            # Same here, yaml file doesn't have sys info
            log.info("Config file doesn't contain any system info")
        else:
            log.info("Loading system options from configuration file")
            if system:
                system = self.update_from_yaml(system, yamloptions)
            else:
                system = self.system_from_yaml(yamloptions)

        if system:
            if not yamloptions:
                print("System is being built only off of the system driver")
            return system
        else:
            log.info("No system specified! Exiting program.")
            # Gracefully exit
            raise ValueError("No system defined!")

    def system_from_yaml(self, system_dict):
        """
        System builder directly from the config YAML file.

        Arguments:
          system_dict (dict): Parsed YAML file as a dictionary, parsed by
            PyYAML by default.

        Returns:
          A modified WESTSystem object as defined in yamlcfg.py with
          the parsed settings from the config file.
        """

        yamlSystem = YAMLSystem()
        print("System building only off of the configuration file")
        # Now for the building of the system from YAML we need to use
        # require for these settings since they are musts.

        # First basic pcoord settings
        ndim = self.config.require(['west', 'system', 'system_options', 'pcoord_ndim'])
        plen = self.config.require(['west', 'system', 'system_options', 'pcoord_len'])
        # Dtype needs to be ran as code from YAML file, document YAML code execution syntax
        # somewhere
        ptype = self.config.require(['west', 'system', 'system_options', 'pcoord_dtype'])
        # Bins
        bins_obj = self.config.require(['west', 'system', 'system_options', 'bins'])
        trgt_cnt = self.config.require(['west', 'system', 'system_options', 'bin_target_counts'])
        # Now add the parsed settings to the system
        mapper = bins_from_yaml_dict(bins_obj)
        setattr(yamlSystem, 'pcoord_ndim', ndim)
        setattr(yamlSystem, 'pcoord_len', plen)
        setattr(yamlSystem, 'pcoord_dtype', ptype)
        setattr(yamlSystem, 'bin_mapper', mapper)
        # Check if the supplied target count object is
        # an iterable or not,

        # This one I designed in a way that you can either
        # directly supply an iterable that has the correct
        # size OR an integer. Anything else will fail.

        # Main issue: For higher bin dimensions this
        # is tough since the bins might not correspond
        # properly to the flattened array you are supplying.

        # I might just scrap the iterable later and only allow
        # integers.
        if hasattr(trgt_cnt, "__iter__"):
            assert len(trgt_cnt) == mapper.nbins, "Count iterable size doesn't match the number of bins"
            trgt_cnt_arr = trgt_cnt
        else:
            assert trgt_cnt == int(trgt_cnt), "Counts are not integer valued, ambiguous input"
            trgt_cnt_arr = np.zeros(mapper.nbins)
            trgt_cnt_arr[:] = trgt_cnt
        setattr(yamlSystem, 'bin_target_counts', trgt_cnt_arr)

        # Attach generic attribute to system
        for attr in system_dict.keys():
            if not hasattr(yamlSystem, attr):
                setattr(yamlSystem, attr, system_dict[attr])

        # Return complete system
        return yamlSystem

    def update_from_yaml(self, init_system, system_dict):
        """
        Updates the system built from the driver with the options
        from the YAML file. For now it overwrites everything specified
        in the system driver.

        Arguments:
          system_dict (dict): Parsed YAML file as a dictionary, parsed by
            PyYAML by default.
          init_system (WESTSystem): System returned by the driver.

        Returns:
          A modified WESTSystem object with settings from the system
          driver and the config YAML file.
        """

        # First we want to overwrite whatever we have from the YAML
        # file.
        print("Updating system with the options from the configuration file")
        for key, value in system_dict.items():
            if key == 'pcoord_ndim':
                self.overwrite_option(init_system, key, value)
            elif key == 'pcoord_len':
                self.overwrite_option(init_system, key, value)
            elif key == 'pcoord_dtype':
                self.overwrite_option(init_system, key, value)
            elif key == "bins":
                self.overwrite_option(init_system, 'bin_mapper', bins_from_yaml_dict(value))
        # Target counts have to be parsed after we have a mapper in
        # place
        try:
            trgt_cnt = system_dict['bin_target_counts']
            if hasattr(trgt_cnt, "__iter__"):
                assert len(trgt_cnt) == init_system.bin_mapper.nbins, "Count iterable size doesn't match the number of bins"
                trgt_cnt_arr = trgt_cnt
            else:
                assert trgt_cnt == int(trgt_cnt), "Counts are not integer valued, ambiguous input"
                trgt_cnt_arr = np.zeros(init_system.bin_mapper.nbins)
                trgt_cnt_arr[:] = int(trgt_cnt)
            self.overwrite_option(init_system, 'bin_target_counts', trgt_cnt_arr)
        except KeyError:
            pass
        # The generic attribute settings added here
        for attr in system_dict.keys():
            if not hasattr(init_system, attr):
                setattr(init_system, attr, system_dict[attr])
        return init_system

    def overwrite_option(self, system, key, value):
        if hasattr(system, key):
            log.info("Overwriting system option: %s" % key)
        setattr(system, key, value)

    def get_system_driver(self):
        if self._system is None:
            self._system = self.new_system_driver()
        return self._system

    def get_work_manager(self):
        return self.work_manager

    def clear_state(self):
        self._sim_manager = None
        self._system = None
        self._data_manager = None
        self._we_driver = None
        self._propagator = None

    propagator = property(get_propagator)
    we_driver = property(get_we_driver)
    system = property(get_system_driver)
    data_manager = property(get_data_manager)
    sim_manager = property(get_sim_manager)
